from cli.utils import auth_headers, download_url, parse_dataset_filter
from zpy.files import to_pathlib_path
from table_logger import TableLogger
import requests
import logging
import json

log = logging.getLogger(__name__)


def create_generated_dataset(name, scene_name, config, url, token):
    """ create a dataset on ragnarok """
    endpoint = f'{url}/api/v1/scenes/'
    params = {'name': scene_name}
    r = requests.get(endpoint, params=params, headers=auth_headers(token))
    if r.status_code != 200:
        log.warning(f'unable to fetch scenes')
        return
    response = json.loads(r.text)
    if response['count'] != 1:
        log.warning(f'unable to find scene with name {scene_name}')
        return
    scene = response['results'][0]
    endpoint = f'{url}/api/v1/generated-data-sets/'
    data = {
        'scene': scene['id'],
        'config': json.dumps(config),
        'name': name
    }
    r = requests.post(endpoint, data=data, headers=auth_headers(token))
    if r.status_code != 201:
        log.warning(f'Unable to create dataset {name} for scene {scene_name} with config {config}')
        return
    log.info(f'created dataset {name} for scene {scene_name} with config {config}')


def filter_dataset(dfilter, url, token):
    """ filter dataset """
    dset = []
    field, pattern, regex = parse_dataset_filter(dfilter)
    endpoint = f'{url}/api/v1/uploaded-data-sets/'
    dset.extend(filter_dataset_url(field, pattern, regex, endpoint, token))
    endpoint = f'{url}/api/v1/generated-data-sets/'
    dset.extend(filter_dataset_url(field, pattern, regex, endpoint, token))
    # TODO figure out a way to have the user ask which set to query
    #endpoint = f'{url}/api/v1/job-data-sets/'
    #dset.extend(filter_dataset_url(field, pattern, regex, endpoint, token))
    return dset


def filter_dataset_url(field, pattern, regex, url, token):
    """ filter generated dataset """
    endpoint = f'{url}?{field}__{pattern}={regex}'
    datasets, names = [], []
    while endpoint is not None:
        r = requests.get(endpoint, headers=auth_headers(token))
        if r.status_code != 200:
            log.warning(f"Unable to filter {endpoint}")
            return []
        response = json.loads(r.text)
        for d in response['results']:
            names.append(d['name'])
            datasets.append(d['id'])
        endpoint = response['next']
    log.info(f'filter <{endpoint}> found {names}')
    return datasets


def create_uploaded_dataset(name, path, url, token):
    """ uploaded a dataset to ragnarok """
    endpoint = f'{url}/api/v1/uploaded-data-sets/'
    data = {'name': name}
    files = {'file': open(path, 'rb')}
    r = requests.post(endpoint, data=data, headers=auth_headers(token), files=files)
    if r.status_code != 201:
        log.warning(f'unable to create dataset {name} from {path}')
        return
    log.info(f'created dataset {name} from {path}')


def fetch_dataset(name, path, dataset_type, url, token):
    """ fetch a dataset from ragnarok """
    endpoint = f'{url}/api/v1/{dataset_type}-data-sets/'
    params = {'name': name}
    r = requests.get(endpoint, params=params, headers=auth_headers(token))
    if r.status_code != 200:
        log.warning(f'Unable to fetch {dataset_type} datasets')
        return
    response = json.loads(r.text)
    if response['count'] != 1:
        log.warning(f'Unable to find {dataset_type} dataset with name "{name}"')
        return
    dataset = response['results'][0]
    endpoint = f"{url}/api/v1/{dataset['dataset_type']}-data-sets/{dataset['id']}/download"
    r = requests.get(endpoint, headers=auth_headers(token))
    if r.status_code != 200:
        log.warning(f"Unable to get download link for dataset {dataset['id']}")
        return
    response = json.loads(r.text)
    name_slug = f"{dataset['name'].replace(' ', '_')}-{dataset['id'][:8]}.zip"
    output_path = to_pathlib_path(path) / name_slug
    download_url(response['redirect_link'], output_path)


def fetch_datasets(endpoint, token):
    """ fetch all datasets in ragnarok """
    u_datasets = fetch_uploaded_datasets(endpoint, token)
    g_datasets = fetch_generated_datasets(endpoint, token)
    j_datasets = fetch_job_datasets(endpoint, token)
    tbl = TableLogger(columns='state,type,name,timestamp,id',default_colwidth=30)
    for d in u_datasets:
        tbl(d['state'], 'UPLOADED', d['name'], d['created_at'], d['id'])
    for d in g_datasets:
        tbl(d['state'], 'GENERATED', d['name'], d['created_at'], d['id'])
    for d in j_datasets:
        tbl(d['state'], 'JOB', d['name'], d['created_at'], d['id'])


def fetch_uploaded_datasets(endpoint, token):
    endpoint = f'{endpoint}/api/v1/uploaded-data-sets/'
    r = requests.get(endpoint, headers=auth_headers(token))
    if r.status_code != 200:
        log.warning('Unable to fetch uploaded datasets')
        return []
    return json.loads(r.text)['results']


def fetch_generated_datasets(endpoint, token):
    endpoint = f'{endpoint}/api/v1/generated-data-sets/'
    r = requests.get(endpoint, headers=auth_headers(token))
    if r.status_code != 200:
        log.warning('Unable to fetch generated datasets')
        return []
    return json.loads(r.text)['results']


def fetch_job_datasets(endpoint, token):
    endpoint = f'{endpoint}/api/v1/job-data-sets/'
    r = requests.get(endpoint, headers=auth_headers(token))
    if r.status_code != 200:
        log.warning('Unable to fetch job datasets')
        return []
    return json.loads(r.text)['results']



"""
    File utilities.
"""
import csv
import json
import logging
import os
import random
import re
import shutil
import subprocess
import sys
import tempfile
import zipfile
from pathlib import Path
from pprint import pformat
from typing import Any, Dict, List, Union

log = logging.getLogger(__name__)


'''
Dictionary of filename extensions and prefix/suffixes

These serve as the master search patterns so update and use
these as much as possible to prevent bugs.

Order matters! More specific regexes at the top and
catch-alls at the bottom.

You can test these out at: https://regex101.com/
'''
IMAGE_REGEX = '.*\.(jpeg|jpg|png|bmp)'
FILE_REGEX = {
    # Images
    'instance segmentation image': 'IMG_[0-9]*.*iseg' + IMAGE_REGEX,
    'class segmentation image': 'IMG_[0-9]*.*cseg' + IMAGE_REGEX,
    'depth image': 'IMG_[0-9]*.*depth' + IMAGE_REGEX,
    'normal image': 'IMG_[0-9]*.*normal' + IMAGE_REGEX,
    'stereo left image': 'IMG_[0-9]*.*stereoL' + IMAGE_REGEX,
    'stereo right image': 'IMG_[0-9]*.*stereoR' + IMAGE_REGEX,
    'rgb image': 'IMG_[0-9]*.*rgb' + IMAGE_REGEX,
    'custom image': 'IMG_[0-9]*' + IMAGE_REGEX,
    'image': IMAGE_REGEX,
    # Annotations
    'zumo annotation': 'ZUMO_META.json',
    'coco annotation': '.*coco.*\.json',
    'annotation': '.*\.(json|xml|yaml)',
}


def dataset_contents(
    path: Union[str, Path],
    filetype_regex: Dict = FILE_REGEX,
) -> Dict:
    """Use regex to search inside a data directory."""
    path = verify_path(path, check_dir=True, make=False)
    contents = {
        'dirs': [],
    }
    for dirpath, _, files in os.walk(path):
        contents['dirs'].append(dirpath)
        for filename in files:
            for name, re_pattern in filetype_regex.items():
                if re.search(re_pattern, filename):
                    if contents.get(name, None) is None:
                        contents[name] = []
                    contents[name].append(os.path.join(dirpath, filename))
                    break
    return contents


def file_is_of_type(filename: Union[str, Path], filetype: str) -> bool:
    """ Check to see if file is of type given by regex."""
    if isinstance(filename, Path):
        filename = str(filename)
    assert FILE_REGEX.get(filetype, None) is not None, \
        f'{filetype} must be in {FILE_REGEX.keys()}'
    if re.search(FILE_REGEX[filetype], filename):
        return True
    return False


def make_rgb_image_name(idx: int, extension: str = '.png') -> str:
    """ Return rgb image name from image id. """
    return 'IMG_%08d_rgb' % idx + extension


def make_custom_image_name(idx: int, name: str, extension: str = '.png') -> str:
    """ Return rgb image name from image id. """
    return 'IMG_%08d_%s' % (idx, name) + extension


def make_cseg_image_name(idx: int, extension: str = '.png') -> str:
    """ Return class segmentation image name from image id. """
    return 'IMG_%08d_cseg' % idx + extension


def make_iseg_image_name(idx: int, extension: str = '.png') -> str:
    """ Return instance segmentation image name from image id. """
    return 'IMG_%08d_iseg' % idx + extension


def make_depth_image_name(idx: int, extension: str = '.png') -> str:
    """ Return instance segmentation image name from image id. """
    return 'IMG_%08d_depth' % idx + extension


def frame_from_image_name(image_name: str) -> int:
    """ Extract frame id from image name. """
    return int(''.join([s for s in image_name if s.isdigit()]))


def replace_index_in_image_name(image_name: str, new_idx : int) -> str:
    """ Replace the id in an image name. """
    # HACK: This will break for image names without 8-digit indices
    return image_name[:4] + '%08d' % new_idx + image_name[12:]


def make_underscore_path(path: Union[str, Path], name: str) -> Path:
    """ Make an underscore path: foo.txt -> foo_new.txt """
    path = to_pathlib_path(path)
    underscore_filename = path.stem + '_' + name + path.suffix
    return path.parent / Path(underscore_filename)


def to_pathlib_path(path: Union[str, Path]) -> Path:
    """Convert string path to pathlib.Path if needed."""
    if not isinstance(path, Path):
        path = Path(os.path.expandvars(path)).resolve()
    return path


def default_temp_path() -> Path:
    """ Temporary directory agnostic to OS """
    return Path(tempfile.gettempdir()) / 'output'


def clean_dir(
    path: Union[str, Path],
    keep_dir: bool = True,
) -> None:
    """Delete everything at the provided directory."""
    path = verify_path(path, make=False, check_dir=True)
    if keep_dir:
        # Delete the contents, but keep the directory
        for _path in path.iterdir():
            try:
                if _path.is_file() or _path.is_symlink():
                    _path.unlink()
                elif _path.is_dir():
                    shutil.rmtree(_path)
            except Exception as e:
                log.warning('Failed to delete %s. Reason: %s' % (_path, e))
    else:
        # Delete everything, including the directory itself
        shutil.rmtree(path)


def pretty_print(_d: Dict) -> str:
    """ Pretty print default formatting """
    return pformat(_d, indent=2, width=120)


def verify_path(
    path: Union[str, Path],
    make: bool = False,
    check_dir: bool = False,
) -> Path:
    """Verifies (or creates) directory at path."""
    path = to_pathlib_path(path)
    if not path.exists():
        log.warning(f'Could not find path at {path}')
        if make:
            log.info(f'Making {path.name} dir at {path}')
            path.mkdir(exist_ok=True, parents=True)
    else:
        log.debug(f'Path found at {path}.')
        if check_dir and not path.is_dir():
            raise ValueError(f'Path at {path} is not a directory.')
    return path


def write_json(path: Union[str, Path], data: Union[Dict, List]) -> None:
    """ Save data to path """
    path = to_pathlib_path(path)
    log.info(f'Writing JSON to file {path}')
    with path.open('w') as f:
        json.dump(data, f, indent=4)


def read_json(path: Union[str, Path]) -> Union[Dict, List]:
    """ Read data from path """
    path = to_pathlib_path(path)
    log.info(f'Reading JSON file at {path}')
    with path.open() as f:
        data = json.load(f)
    return data


def write_csv(
    path: Union[str, Path],
    data: List[List[Any]],
    delimiter: str = ',',
    **kwargs
) -> None:
    """ Save data to path """
    path = to_pathlib_path(path)
    log.info(f'Writing CSV to file {path}')
    with path.open('w') as f:
        writer = csv.writer(f, delimiter=delimiter, **kwargs)
        writer.writerows(data)


def read_csv(
    path: Union[str, Path],
    delimiter: str = ',',
    **kwargs
) -> List[List[Any]]:
    """ Read data from path """
    path = to_pathlib_path(path)
    log.info(f'Reading CSV file at {path}')
    data = []
    with path.open() as f:
        for row in csv.reader(f, delimiter=delimiter, **kwargs):
            data.append(row)
    return data


def sample(things: List, sample_size: int = None) -> List:
    """ Return a sample of things. """
    random_sample_size = len(things)
    if sample_size is not None:
        random_sample_size = min(sample_size, len(things))
    if random_sample_size == len(things):
        sample_images = things
    else:
        sample_images = random.sample(things, random_sample_size)
    return sample_images


def filecopy(
    src_dir: Union[str, Path] = None,
    dst_dir: Union[str, Path] = None,
    src_name: str = None,
    dst_name: str = None,
) -> None:
    """ Copy over a file. """
    src_dir = verify_path(src_dir, check_dir=True)
    dst_dir = verify_path(dst_dir, check_dir=True)
    src = src_dir / src_name
    dst = dst_dir / dst_name
    src = verify_path(src)
    dst = verify_path(dst)
    log.debug(f'Copying over file from {src} to {dst}')
    shutil.copy(src, dst)


def open_folder_in_explorer(
    path: Union[str, Path],
    make: bool = False,
) -> None:
    """ Opens a directory in the fileexplorer of your OS. """
    path = verify_path(path, check_dir=True, make=make)
    if sys.platform.startswith('darwin'):
        subprocess.call(('open', path))
    elif os.name == 'nt':
        os.startfile(path)
    elif os.name == 'posix':
        subprocess.call(('xdg-open', path))


def remove_files_with_suffix(
    path: Union[str, Path],
    exts: List[str],
) -> None:
    """Remove file in a path with certain extension """
    path = verify_path(path, check_dir=True)
    for _path in path.glob("*"):
        if _path.suffix in exts:
            log.info(f'Removing file at {_path}')
            _path.unlink()


def unzip_file(
    zip_path: Union[str, Path],
    out_path: Union[str, Path],
) -> None:
    """ Unzip a file at a local path. """
    log.info(f'Unzipping {zip_path} to {out_path}...')
    zip_path = verify_path(zip_path)
    out_path = verify_path(out_path, check_dir=True)
    if not zip_path.suffix == '.zip':
        raise ValueError(f'{zip_path} is not a zip file')

    zf = zipfile.ZipFile(str(zip_path))
    zipped_size_mb = round(
        sum([i.compress_size for i in zf.infolist()]) / 1024 / 1024)
    unzipped_size_mb = round(
        sum([i.file_size for i in zf.infolist()]) / 1024 / 1024)
    log.info(f'Compressed: {zipped_size_mb}MB, actual: {unzipped_size_mb}MB.')
    zf.extractall(out_path)
    log.info(f'Done extracting to {out_path}.')


def zip_file(
    in_path: Union[str, Path],
    zip_path: Union[str, Path],
) -> None:
    """ Zip a file at a local path. """
    log.info(f'Zipping {in_path} to {zip_path}...')
    in_path = verify_path(in_path)
    zip_path = verify_path(zip_path)
    if not zip_path.suffix == '.zip':
        raise ValueError(f'{zip_path} is not a zip file')

    shutil.make_archive(
        base_name=zip_path.parent / zip_path.stem,
        format='zip',
        root_dir=in_path)
    log.info(f'Done zipping to {zip_path}.')

    zf = zipfile.ZipFile(str(zip_path))
    zipped_size_mb = round(
        sum([i.compress_size for i in zf.infolist()]) / 1024 / 1024)
    unzipped_size_mb = round(
        sum([i.file_size for i in zf.infolist()]) / 1024 / 1024)
    log.info(f'Compressed: {zipped_size_mb}MB, actual: {unzipped_size_mb}MB.')



"""
    ZUMO dataset format.
"""
import logging
from pathlib import Path
from typing import Dict, Union

import gin

import zpy
from zpy.output import Output
from zpy.saver_image import ImageSaver

log = logging.getLogger(__name__)


class ZUMOParseError(Exception):
    """ Invalid ZUMO Annotation found when parsing data contents. """
    pass


@gin.configurable
class OutputZUMO(Output):
    """Holds the logic for outputting ZUMO annotations to file."""

    ANNOTATION_FILENAME = Path('ZUMO_META.json')

    def output_annotations(self,
                           annotation_path: Union[str, Path] = None,
                           ):
        """ Output ZUMO_META annotations.

        The ZUMO format is meant to be as close to
        a serialized version of this class as possbile.

        """
        log.info('output zumo annotations...')
        zumo_dict = {
            'metadata': self.saver.metadata,
            'categories': self.saver.categories,
            'images': self.saver.images,
            'annotations': self.saver.annotations,
        }
        # Get the correct annotation path
        if annotation_path is not None:
            annotation_path = annotation_path
        elif self.saver.annotation_path is None:
            annotation_path = self.saver.output_dir / self.ANNOTATION_FILENAME
        else:
            annotation_path = self.saver.annotation_path
        # Write out annotations to file
        zpy.files.write_json(annotation_path, zumo_dict)
        # Verify annotations
        parse_zumo_annotations(
            annotation_file=annotation_path, data_dir=self.saver.output_dir)


@gin.configurable
def parse_zumo_annotations(
    annotation_file: Union[str, Path],
    data_dir: Union[str, Path] = None,
    output_saver: bool = False,
) -> None:
    """ Parse Zumo annotations. """
    log.info(f'Parsing ZUMO annotations at {annotation_file}...')

    # Check annotation file path
    annotation_file = zpy.files.verify_path(annotation_file)
    if data_dir is not None:
        data_dir = zpy.files.verify_path(data_dir, check_dir=True)
    else:
        # If no data_dir, assume annotation file is in the root folder.
        data_dir = annotation_file.parent

    zumo_metadata = zpy.files.read_json(annotation_file)
    images = zumo_metadata['images']
    if len(images.keys()) == 0:
        raise ZUMOParseError(f'no images found in {annotation_file}')
    categories = zumo_metadata['categories']
    if len(categories.keys()) == 0:
        raise ZUMOParseError(f'no categories found in {annotation_file}')
    annotations = zumo_metadata['annotations']
    if len(annotations) == 0:
        raise ZUMOParseError(
            f'no annotations found in {annotation_file}')
    log.info(
        f'images:{len(images)} categories:{len(categories)} annotations:{len(annotations)}')

    # Optionally output a saver object.
    if output_saver:
        saver = ImageSaver(output_dir=data_dir,
                           annotation_path=annotation_file,
                           description=zumo_metadata['metadata']['description'],
                           clean_dir=False,
                           )

    # Check Images
    log.info('Parsing images...')
    img_ids = []
    for image_id, img in images.items():
        # HACK: JSON will convert int keys to str, so undo that here
        image_id = int(image_id)
        # Image ID
        if not image_id == img['id']:
            raise ZUMOParseError(
                f"image id {image_id} does not match image dict key {img['id']}")
        if not isinstance(image_id, int):
            raise ZUMOParseError(f'image id {image_id} must be int.')
        if image_id in img_ids:
            raise ZUMOParseError(f'image id {image_id} already used.')
        img_ids.append(image_id)
        if image_id < 0:
            raise ZUMOParseError(f'invalid image id {image_id}')
        # Frame
        frame = img.get('frame', None)
        if frame is not None:
            if not isinstance(frame, int):
                raise ZUMOParseError(f'frame {frame} must be int.')
            if image_id < 0:
                raise ZUMOParseError(f'invalid image frame {frame}')
        # Height and Width
        height, width = img['height'], img['width']
        if not isinstance(height, int):
            raise ZUMOParseError(f'height {height} must be int.')
        if not isinstance(width, int):
            raise ZUMOParseError(f'width {width} must be int.')
        if height <= 0 or width <= 0:
            raise ZUMOParseError(
                f'width and height h:{height} w:{width} must be > 0')
        # Name
        name = img.get('name', None)
        if name is not None:
            if not isinstance(name, str):
                raise ZUMOParseError(f'name {name} must be str.')
            if frame is not None and \
                (not zpy.files.frame_from_image_name(name) == frame) and \
                    (not zpy.files.frame_from_image_name(name) == image_id):
                raise ZUMOParseError(f'name {name} does not correspond to'
                                     f' frame {frame} or image_id {image_id}.')
        # Output path
        output_path = img.get('output_path', None)
        if output_path is not None:
            if not isinstance(output_path, str):
                raise ZUMOParseError(f'output_path {output_path} must be str.')
            if not Path(output_path).exists():
                raise ZUMOParseError(
                    f'output_path {output_path} does not exist')
        # Save each image to ImageSaver object
        if output_saver:
            saver.images[image_id] = img

    # Check Categories
    log.info('Parsing categories...')
    cat_ids = []
    cat_names = []
    for category_id, category in categories.items():
        # Category Name
        category_name = category['name']
        if not isinstance(category_name, str):
            raise ZUMOParseError(
                f'category_name {category_name} must be str.')
        if category_name in cat_names:
            raise ZUMOParseError(
                f'category_name {category_name} already used')
        cat_names.append(category_name)
        # HACK: JSON will convert int keys to str, so undo that here
        category_id = int(category_id)
        # Category ID
        if not category_id == category['id']:
            raise ZUMOParseError(
                f"category_id {category_id} does not match category dict key {category['id']}")
        if not isinstance(image_id, int):
            raise ZUMOParseError(
                f'category_id {category_id} must be int.')
        if category_id in cat_ids:
            raise ZUMOParseError(
                f'category id {category_id} already used')
        cat_ids.append(category_id)
        # Supercategories
        if category.get('supercategory', None) is not None:
            pass
        if category.get('supercategories', None) is not None:
            pass
        # Subcategories
        if category.get('subcategory', None) is not None:
            pass
        if category.get('subcategories', None) is not None:
            pass
        # Keypoints
        if category.get("keypoints", None) is not None:
            keypoints = category['keypoints']
            log.info(f"{len(keypoints)} keypoints:{keypoints}")
            if category.get('skeleton', None) is None:
                raise ZUMOParseError(
                    f'skeleton must be present with {keypoints}')
        # Save each category to ImageSaver object
        if output_saver:
            saver.categories[category_id] = category

    # Check Annotations
    log.info('Parsing annotations...')
    ann_ids = []
    for annotation in annotations:
        # IDs
        image_id, category_id, annotation_id = annotation[
            'image_id'], annotation['category_id'], annotation['id']
        if image_id not in img_ids:
            raise ZUMOParseError(
                f'annotation image id {image_id} not in {img_ids}')
        if category_id not in cat_ids:
            raise ZUMOParseError(
                f'annotation category id {category_id} not in {cat_ids}')
        if annotation_id in ann_ids:
            raise ZUMOParseError(
                f'annotation id {annotation_id} already used')
        ann_ids.append(annotation_id)

        # Bounding Boxes
        bbox = annotation.get('bbox', None)
        if bbox is not None:
            pass

        # Keypoints
        keypoints = annotation.get('num_keypoints', None)
        if keypoints is not None:
            if "keypoints_xyv" in annotation:
                if len(annotation["keypoints_xyv"]) != int(annotation["num_keypoints"])*3:
                    raise ZUMOParseError(
                        'keypoints_xyv not correct size {len(keypoints)}')
            if "keypoints_xyz" in annotation:
                if len(annotation["keypoints_xyz"]) != int(annotation["num_keypoints"])*3:
                    raise ZUMOParseError(
                        'keypoints_xyz not correct size {len(keypoints)}')

        # Save each annotation to ImageSaver object
        if output_saver:
            saver.annotations.append(annotation)

    if output_saver:
        return saver
    else:
        return None



"""
    Node utilities.
"""
import logging

import bpy
import mathutils
from typing import Tuple

import gin
import zpy

log = logging.getLogger(__name__)


@gin.configurable
def get_or_make(
    name: str,
    node_type: str,
    tree: bpy.types.NodeTree,
    label_tag: str = '(zpy) ',
    pos: Tuple[float] = None,
) -> bpy.types.Node:
    """ Verify existence or create a node.

    Args:
        name (str): Name of the node.
        node_type (str): Node type e.g. "ShaderNodeBackground"
        tree (bpy.types.NodeTree): Node tree where this node will be added.
        label_tag (str, optional): Node label will include this tag to make it easier to identify within Blender. Defaults to '(zpy) '.
        pos (Tuple[float], optional): Location of the node in node editor. Defaults to None.

    Returns:
        bpy.types.Node: The newly created (or already existing) node.
    """
    node = tree.nodes.get(name, None)
    if node is None:
        node = tree.nodes.new(node_type)
        node.name = name
    node.label = f'{label_tag}{name}'
    node.bl_description = 'This node has been created and/or modified by zpy'
    if pos is not None:
        node.location = pos
    return node



"""
    Logging utilities.
"""
import logging
from typing import List

log = logging.getLogger(__name__)


def set_log_levels(
    level: str = None,
    modules: List[str] = [
        'zpy',
        # Addon modules
        'zpy_addon',
        'bpy.zpy_addon'
        'neuralyzer',
    ]
) -> None:
    """ Set logger levels for all zpy modules. """
    if level is None:
        log_level = logging.INFO
    elif level == 'info':
        log_level = logging.INFO
    elif level == 'debug':
        log_level = logging.DEBUG
    elif level == 'warning':
        log_level = logging.WARNING
    else:
        log.warning(f'Invalid log level {level}')
        return
    log.warning(f'Setting log level to {log_level} ({level})')
    logging.basicConfig(
        level=log_level,
        format='%(asctime)s: %(levelname)s %(filename)s] %(message)s')
    for logger_name in modules:
        try:
            logging.getLogger(logger_name).setLevel(log_level)
        except:
            pass


def linebreaker_log(message: str, line_length: int = 80):
    """ Good looking line-breaker log message. """
    # Clip the message
    message = message[:line_length]
    whitespace = ' ' * int((line_length - len(message)) / 2)
    # La piece de resistance
    log.info('-'*line_length)
    log.info(f'{whitespace}{message.upper()}{whitespace}')
    log.info('-'*line_length)



"""
    Video version of Saver object.
"""
import logging
from datetime import date
from pathlib import Path
from typing import Dict, List, Tuple, Union
from datetime import timedelta

import gin
import numpy as np

import zpy

log = logging.getLogger(__name__)


@gin.configurable
class VideoSaver(zpy.saver.Saver):
    """Holds the logic for saving video annotations at runtime."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.videos = {}
        self.video_name_to_id = {}

    @gin.configurable
    def add_video(self,
                  name: str = 'default video',
                  style: str = 'default',
                  output_path: Union[str, Path] = '/tmp/test.avi',
                  width: int = 0,
                  height: int = 0,
                  length: timedelta = 0,
                  zero_indexed: bool = True,
                  **kwargs,
                  ) -> Dict:
        """ Add image to save object. """
        video = {
            'name': name,
            'style': style,
            'output_path': str(output_path),
            'relative_path': str(Path(output_path).relative_to(self.output_dir)),
            'width': width,
            'height': height,
            'length': length,
        }
        video.update(**kwargs)
        video['id'] = len(self.videos.keys())
        video['id'] += 0 if zero_indexed else 1
        log.debug(f'Adding video: {zpy.files.pretty_print(video)}')
        self.videos[video['id']] = video
        self.video_name_to_id[name] = video['id']
        return video

    @gin.configurable
    def add_annotation(self,
                       *args,
                       video: str = 'default video',
                       **kwargs,
                       ) -> Dict:
        """ Add annotation. """
        annotation = super().add_annotation(*args, **kwargs)
        video_id = self.video_name_to_id.get(video, None)
        assert video_id is not None, f'Could not find id for video {video}'
        annotation['video_id'] = video_id
        annotation.update(**kwargs)
        self.annotations.append(annotation)
        return annotation

    @gin.configurable
    def output_meta_analysis(self):
        """ Perform a full meta analysis.  """
        log.warning('TODO: implement meta-analysis for video datasets')



"""
    HDRI utilities. You can find lots of free HDRIs at https://hdrihaven.com/
"""
import logging
import random
import math
from pathlib import Path
from typing import Dict, List, Tuple, Union

import bpy
import gin
import mathutils

import zpy

log = logging.getLogger(__name__)


@gin.configurable
def load_hdri(
    path: Union[str, Path],
    scale: Tuple[float] = (1.0, 1.0, 1.0),
    random_z_rot: bool = True,
) -> None:
    """ Load an HDRI from path. """
    scene = zpy.blender.verify_blender_scene()
    scene.world.use_nodes = True
    tree = scene.world.node_tree
    out_node = zpy.nodes.get_or_make(
        'World Output', 'ShaderNodeOutputWorld', tree, pos=(0, 0))
    bg_node = zpy.nodes.get_or_make(
        'Background', 'ShaderNodeBackground', tree, pos=(-150, 0))
    env_node = zpy.nodes.get_or_make(
        'Environment Texture', 'ShaderNodeTexEnvironment', tree, pos=(-400, 0))
    log.info(f'Loading HDRI at {path}')
    path = zpy.files.verify_path(path, make=False)
    env_node.image = bpy.data.images.load(str(path))
    env_node.texture_mapping.scale = mathutils.Vector(scale)
    world_rot_node = zpy.nodes.get_or_make(
        'World Rotation', 'ShaderNodeVectorRotate', tree, pos=(-550, 0))
    world_rot_node.rotation_type = 'Z_AXIS'
    if random_z_rot:
        world_rotation = random.uniform(0, math.pi)
        log.debug(f'Rotating HDRI randomly along Z axis to {world_rotation}')
        world_rot_node.inputs['Angle'].default_value = world_rotation
    texcoord_node = zpy.nodes.get_or_make(
        'Texture Coordinate', 'ShaderNodeTexCoord', tree, pos=(-730, 0))
    # Link all the nodes together
    tree.links.new(out_node.inputs[0], bg_node.outputs[0])
    tree.links.new(bg_node.inputs[0], env_node.outputs[0])
    tree.links.new(env_node.inputs[0], world_rot_node.outputs[0])
    tree.links.new(world_rot_node.inputs[0], texcoord_node.outputs[0])


@gin.configurable
def random_hdri(
    hdri_dir: Union[str, Path] = '$ASSETS/lib/hdris/1k',
    apply_to_scene: bool = True,
) -> Path:
    """ Generate a random HDRI from an asset path. """
    hdri_dir = zpy.files.verify_path(hdri_dir, make=False, check_dir=True)
    # Create list of HDRIs in directory
    hdris = []
    for _path in hdri_dir.iterdir():
        if _path.is_file() and _path.suffix == '.hdri':
            hdris.append(_path)
    hdri_path = random.choice(hdris)
    log.info(f'Found {len(hdris)} HDRIs, randomly chose {hdri_path.stem}')
    if apply_to_scene:
        load_hdri(hdri_path)
    return hdri_path



"""
    Blender Addons require a special __init__ file.
"""
import importlib
import logging
import subprocess
import sys
from pathlib import Path

import bpy

log = logging.getLogger(__name__)

bl_info = {
    "name": "zpy",
    "author": "Zumo Labs",
    # TODO: Keep up to date with $ZPY_VERSION
    "version": (1, 0, 0),
    # TODO: Keep up to date with $BLENDER_VERSION
    "blender": (2, 92, 0),
    "location": "View3D > Properties > zpy",
    "description": "Synthetic data creation tools for Blender.",
    "warning": "",
    "doc_url": "https://github.com/ZumoLabs/zpy/tree/main/README.md",
    "category": "3D View",
}


def install_pip_depenencies():
    """ Install pip dependencies required by zpy addon."""
    try:
        log.info('Installing zpy and dependencies...')
        # Upgrade pip with Blender's python
        pip_install = [sys.executable, "-m", "pip", "install"]
        subprocess.run(pip_install + ["--upgrade", "pip"], check=True)
        # Install zpy through PyPI into Blender's python site-package dir
        pkg_path = Path(sys.executable).parent.parent / \
            'lib' / 'site-packages' / 'zpy'
        subprocess.run(pip_install + ['zpy-zumo', '--target', str(pkg_path)],
                       check=True)
    except Exception as e:
        log.warning(f'Could not install ZPY and dependencies: {e}')


try:
    import zpy
except ModuleNotFoundError as e:
    log.exception('No zpy module found, please follow developer ' + \
        'install instructions at https://github.com/ZumoLabs/zpy#install')
    # TODO: Automatic installation of pip dependencies
    #       waiting on https://developer.blender.org/T71420
    # install_pip_depenencies()
    # import zpy


if "bpy" in locals():
    log.warning('Reloading zpy_addon files.')
    from . import export_panel
    from . import output_panel
    from . import execute_panel
    from . import script_panel
    from . import segment_panel
    importlib.reload(export_panel)
    importlib.reload(output_panel)
    importlib.reload(execute_panel)
    importlib.reload(script_panel)
    importlib.reload(segment_panel)
    importlib.reload(zpy)

classes = (
    # Properties
    segment_panel.CategoryProperties,
    segment_panel.SegmentableProperties,
    # Object operators
    segment_panel.SegmentInstanceSingle,
    segment_panel.SegmentInstanceMany,
    segment_panel.ResetSegData,
    # Scene operators
    segment_panel.VisualizeInstance,
    segment_panel.VisualizeCategory,
    segment_panel.CategoriesFromText,
    segment_panel.CategoriesFromZUMOJSON,
    output_panel.OpenOutputDirOperator,
    output_panel.CleanOutputDirOperator,
    execute_panel.RenderOperator,
    execute_panel.RunOperator,
    export_panel.ExportOperator,
    export_panel.OpenExportDirOperator,
    export_panel.CleanUpDirOperator,
    script_panel.LoadGinConfigOperator,
    script_panel.PushGinConfigOperator,
    script_panel.LoadRunpyOperator,
    script_panel.PushRunpyOperator,
    # Panels
    output_panel.SCENE_PT_OutputPanel,
    execute_panel.SCENE_PT_ExecutePanel,
    segment_panel.SCENE_PT_SegmentPanel,
    script_panel.SCENE_PT_ScriptPanel,
    export_panel.SCENE_PT_ExportPanel,
)


def register():
    """ Register any classes and properties. """
    for cls in classes:
        try:
            log.info(f'Registering class {cls.__name__}')
            bpy.utils.register_class(cls)
        except Exception as e:
            log.warning(f'Exception when registering {cls.__name__}: {e}')
    segment_panel.registerObjectProperties()
    segment_panel.registerSceneProperties()
    output_panel.registerSceneProperties()
    export_panel.registerSceneProperties()
    script_panel.registerSceneProperties()


def unregister():
    """ Unregister any classes and properties. """
    for cls in classes:
        try:
            log.info(f'Un-registering class {cls.__name__}')
            bpy.utils.unregister_class(cls)
        except Exception as e:
            log.warning(f'Exception when un-registering {cls.__name__}: {e}')


if __name__ == "__main__":
    register()



"""
    Utilities for color.
"""
import logging
import random
from pathlib import Path
from typing import List, Tuple, Union

import zpy

log = logging.getLogger(__name__)

# Random colors are loaded on module import
RANDOM_COLOR_IDX = 1
COLORS_FILE = 'colors.json'
COLORS = None


def reset():
    """ Load colors from file and reset random idx. """
    global COLORS, RANDOM_COLOR_IDX
    _path = Path(__file__).parent / COLORS_FILE
    zpy.files.verify_path(_path)
    COLORS = zpy.files.read_json(_path)
    RANDOM_COLOR_IDX = 1


def hex_to_irgb(hex_value: str) -> Tuple[int]:
    """ Convert hex value to integer rgb (0 to 255). """
    hex_value = int(hex_value[1:], 16)
    b = (hex_value & 0xFF)
    g = ((hex_value >> 8) & 0xFF)
    r = ((hex_value >> 16) & 0xFF)
    return r, g, b


def frgb_to_frgba(rgb: Tuple[float], a=1.0) -> Tuple[float]:
    """ Convert 3-channel RGB to 4-channel RGBA. """
    _val = *rgb, a
    return _val


def hex_to_frgb(hex_value: str) -> Tuple[float]:
    """ Convert hex value to float rgb (0 to 1). """
    return irgb_to_frgb(hex_to_irgb(hex_value))


def irgb_to_frgb(irgb: Tuple[float]) -> Tuple[float]:
    """ Convert integer rgb (0 to 255) to float rgb (0 to 1). """
    r, g, b = irgb
    return r / 255.0, g / 255.0, b / 255.0


def irgb_to_hex(irgb: Tuple[float]) -> str:
    """ Convert integer rgb (0 to 255) to hex. """
    r, g, b = irgb
    return '#%02x%02x%02x' % (r, g, b)


def frgb_to_irgb(frgb: Tuple[float]) -> Tuple[int]:
    """ Convert float rgb (0 to 1) to integer rgb (0 to 255). """
    r, g, b = frgb
    return int(r * 255.0), int(g * 255.0), int(b * 255.0)


def frgb_to_hex(frgb: Tuple[float]) -> str:
    """ Convert float rgb (0 to 1) to hex. """
    return irgb_to_hex(frgb_to_irgb(frgb))


def frgb_to_srgba(frgb: Tuple[float], a=1.0) -> Tuple[float]:
    """ Convert float rgb (0 to 1) to the gamma-corrected sRGBA float (0 to 1). """
    return frgb[0]**(1/2.2), frgb[1]**(1/2.2), frgb[2]**(1/2.2), a


def frgb_to_srgb(frgb: Tuple[float]) -> Tuple[float]:
    """ Convert float rgb (0 to 1) to the gamma-corrected sRGB float (0 to 1). """
    return frgb[0]**(1/2.2), frgb[1]**(1/2.2), frgb[2]**(1/2.2)


def _output_style(name: str, hex: str, output_style: str) -> Union[Tuple[float, int, str], str]:
    """ Convert hex to an output style. """
    if output_style == 'frgb':
        return hex_to_frgb(hex)
    if output_style == 'frgba':
        return frgb_to_frgba(hex_to_frgb(hex))
    elif output_style == 'irgb':
        return hex_to_irgb(hex)
    elif output_style == 'hex':
        return hex
    elif output_style == 'name_irgb':
        return name, hex_to_irgb(hex)
    elif output_style == 'name_frgb':
        return name, hex_to_frgb(hex)
    elif output_style == 'name_frgba':
        return name, frgb_to_frgba(hex_to_frgb(hex))
    else:
        raise ValueError('Color must be frgb, irgb, or hex.')


def default_color(output_style: str = 'frgb') -> Union[Tuple[float, int, str], str]:
    """ Default color. """
    global COLORS
    if COLORS is None:
        reset()
    _name = COLORS[0]['name']
    _hex = COLORS[0]['hex']
    log.debug(
        f'Default color chosen is {_name} - {_hex} - {hex_to_frgb(_hex)}')
    return _output_style(_name, _hex, output_style=output_style)


def random_color(output_style: str = 'frgb') -> Union[Tuple[float, int, str], str]:
    """ Random color.

    This will go through a pre-baked list every time, 
    to prevent different random seeds from changing the
    color for a category.

    """
    global RANDOM_COLOR_IDX, COLORS
    if COLORS is None:
        reset()
    _name = COLORS[RANDOM_COLOR_IDX]['name']
    _hex = COLORS[RANDOM_COLOR_IDX]['hex']
    # Update global color idx
    RANDOM_COLOR_IDX += 1
    if RANDOM_COLOR_IDX > len(COLORS):
        log.error(f'Ran out of unique colors!')
    log.debug(f'Random color chosen is {_name} - {_hex} - {hex_to_frgb(_hex)}')
    return _output_style(_name, _hex, output_style=output_style)


def closest_color(color: Tuple[float],
                  colors: List[Tuple[float]],
                  max_dist: float = 0.01,
                  ) -> Union[None, Tuple[float]]:
    """ Get the index of the closest color in a list to the input color. """
    min_dist = 3.0
    nearest_idx = 0
    for i, _color in enumerate(colors):
        dist = (color[0] - _color[0])**2 + \
            (color[1] - _color[1])**2 + \
            (color[2] - _color[2])**2
        if dist < min_dist:
            min_dist = dist
            nearest_idx = i
    if min_dist > max_dist:
        log.debug(
            f'No color close enough w/ maxmimum distance of {max_dist}')
        return None
    return colors[nearest_idx]



"""
    Base Saver object for storing annotations, categories, etc during scene runtime.
"""
import logging
from datetime import date
from pathlib import Path
from typing import Dict, List, Tuple, Union

import gin
import numpy as np

import zpy

log = logging.getLogger(__name__)


@gin.configurable
class Saver:
    """ Stores annotations and categories throughout a run script.

    Provides functions for outputing annotation files and meta files.

    Raises:
        ValueError: Incorrect function arguments.

    Returns:
        Saver: The Saver object.
    """

    # Names for annotation files, folders, configs, datasheets, etc
    HIDDEN_METAFOLDER_FILENAME = Path('.zumometa')
    HIDDEN_DATASHEET_FILENAME = Path('ZUMO_DATASHEET.txt')
    GIN_CONFIG_FILENAME = Path('config.gin')

    DATETIME_FORMAT = '20%y%m%d_%H%M%S'
    DATETIME_YEAR_FORMAT = '20%y'

    def __init__(self,
                 output_dir: Union[str, Path] = None,
                 annotation_path: Union[str, Path] = None,
                 description: str = 'Description of dataset.',
                 clean_dir: bool = True,
                 ):
        """ Creates a Saver object.

        Args:
            output_dir (Union[str, Path], optional): Directory where files will be dumped.
            annotation_path (Union[str, Path], optional): Path where annotation file will be dumped.
            description (str, optional): A couple sentences describing the dataset. Defaults to 'Description of dataset.'.
            clean_dir (bool, optional): Whether to empty/clean the output directory on object creation. Defaults to True.
        """
        # the output dir
        if output_dir is None:
            output_dir = zpy.files.default_temp_path()
        self.output_dir = zpy.files.verify_path(
            output_dir, make=True, check_dir=True)
        log.debug(f'Saver output directory at {output_dir}')
        if clean_dir:
            zpy.files.clean_dir(self.output_dir)
        # Annotation files can be optionally written out to a different dir
        if annotation_path is None:
            self.annotation_path = annotation_path
        else:
            self.annotation_path = zpy.files.verify_path(annotation_path)
            log.debug(f'Saver annotation path at {annotation_path}')
        # Very similar keys to COCO-style
        self.metadata = {
            'description': description,
            'contributor': 'Zumo Labs',
            'url': 'zumolabs.ai',
            'year': date.today().strftime(self.DATETIME_YEAR_FORMAT),
            'date_created': date.today().strftime(self.DATETIME_FORMAT),
            'save_path': str(self.output_dir),
        }
        self.categories = {}
        self.annotations = []
        # Reverse-lookup dictionaries for name/color to id
        self.category_name_to_id = {}

    @gin.configurable
    def add_annotation(self,
                       category: str = None,
                       subcategory: str = None,
                       subcategory_zero_indexed: bool = True,
                       **kwargs,
                       ) -> Dict:
        """ Add a new annotation to the Saver object.

        Pass any additional keys you want in the annotation dict as kwargs.

        Args:
            category (str, optional): The category that this annotation belongs to. Defaults to None.
            subcategory (str, optional): The sub-category that this annotation belongs to. Defaults to None.
            subcategory_zero_indexed (bool, optional): Whether sub-categories are zero indexed. Defaults to True.

        Returns:
            Dict: The annotation dictionary.
        """
        annotation = {'id': len(self.annotations)}
        if category is not None:
            category_id = self.category_name_to_id.get(category, None)
            assert category_id is not None, f'Could not find id for category {category}'
            self.categories[category_id]['count'] += 1
            annotation['category_id'] = category_id
        if subcategory is not None:
            subcategory_id = self.categories[category_id]['subcategories'].index(
                subcategory)
            self.categories[category_id]['subcategory_count'][subcategory_id] += 1
            subcategory_id += 0 if subcategory_zero_indexed else 1
            annotation['subcategory_id'] = subcategory_id
        return annotation

    @gin.configurable
    def add_category(self,
                     name: str = 'default',
                     supercategories: List[str] = None,
                     subcategories: List[str] = None,
                     color:  Tuple[float] = (0., 0., 0.),
                     zero_indexed: bool = True,
                     **kwargs,
                     ) -> Dict:
        """ Add a new category (also known as classes) to the Saver object.

        Pass any additional keys you want in the category dict as kwargs.

        Args:
            name (str, optional): Name of the category. Defaults to 'default'.
            supercategories (List[str], optional): Names of any supercategories. Defaults to None.
            subcategories (List[str], optional): Names of any subcategories. Defaults to None.
            color (Tuple[float], optional): Color of the category in segmentation images. Defaults to (0., 0., 0.).
            zero_indexed (bool, optional): Whether categories are zero-indexed. Defaults to True.

        Returns:
            Dict: The category dictionary.
        """
        # Default for super- and sub- categories is empty list
        supercategories = supercategories or []
        subcategories = subcategories or []
        category = {
            'name': name,
            'supercategories': supercategories,
            'subcategories': subcategories,
            'color': color,
            'count': 0,
            'subcategory_count': [0] * len(subcategories),
        }
        category.update(**kwargs)
        category['id'] = len(self.categories.keys())
        category['id'] += 0 if zero_indexed else 1
        log.debug(f'Adding category: {zpy.files.pretty_print(category)}')
        self.categories[category['id']] = category
        self.category_name_to_id[name] = category['id']
        return category

    @gin.configurable
    def remap_filter_categories(self,
                                category_remap: Dict = None,
                                ) -> None:
        """Re-map the categories (name and id correspondence).

        This will also filter out any categories not in the category_remap dict.

        Args:
            category_remap (Dict, optional): Mapping of categorie names to id in {id : name}. Defaults to None.

        Raises:
            ValueError: Incorrect format for category remap dictionary.
        """
        if category_remap is None:
            log.warning(
                'Attempted to remap categories with no category remap.')
            return

        # Intermediate variables for organization
        category_remap_ids = []
        category_remap_name_to_id = {}
        category_remap_old_id_to_new_id = {}

        # Check for duplicates and typing
        for _id, _name in category_remap.items():
            try:
                _id = int(_id)
                _name = str(_name)
            except ValueError:
                raise ValueError('Category remap must be {int : string}')
            if _name in category_remap_name_to_id:
                raise ValueError(f'Duplicate category name in remap: {_name}')
            if _id in category_remap_ids:
                raise ValueError(f'Duplicate category id in remap: {_id}')
            category_remap_ids.append(_id)
            category_remap_name_to_id[_name] = _id

        # Make sure the category names all exist in current categories
        category_names = [c['name'] for c in self.categories.values()]
        for category_name in category_remap_name_to_id:
            assert category_name in category_names, \
                f'Could not find category {category_name} in dataset when remap-ing'

        # Go through all of the current categories
        new_categories = {}
        for old_id, category in self.categories.items():
            # Matching is done using the name
            if category['name'] in category_remap_name_to_id:
                new_id = category_remap_name_to_id[category['name']]
                category_remap_old_id_to_new_id[old_id] = new_id
                category['id'] = new_id
                new_categories[new_id] = category
        # Overwrite the old categories
        self.categories = new_categories

        # Go through all of the current annotations
        new_annotations = []
        # Replace the category_id in all annotations
        for annotation in self.annotations:
            if annotation['category_id'] in category_remap_old_id_to_new_id:
                new_id = category_remap_old_id_to_new_id[annotation['category_id']]
                annotation['category_id'] = new_id
                new_annotations.append(annotation)
        # Overwrite the old annotations
        self.annotations = new_annotations

    def output_gin_config(self):
        """ Output the full gin config. """
        gin_config_filepath = self.output_dir / self.GIN_CONFIG_FILENAME
        log.info(f'Writing out gin config to {gin_config_filepath}')
        with open(gin_config_filepath, "w") as f:
            f.write(gin.operative_config_str())

    @staticmethod
    def write_datasheet(
        datasheet_path: str = None,
        info: Dict = None
    ):
        """ Writes datasheet dict to file.

        Args:
            datasheet_path (str, optional): Path where datasheet will be written.
            info (Dict, optional): Information to include in datasheet.

        """
        with datasheet_path.open('w') as f:
            for k, v in info.items():
                f.write(f'{k},{v}\n')

    @staticmethod
    def clip_coordinate_list(
        annotation: List[Union[int, float]] = None,
        height: Union[int, float] = None,
        width: Union[int, float] = None,
        normalized: bool = False,
    ) -> List[Union[int, float]]:
        """ Clip a list of pixel coordinates (e.g. segmentation polygon).

        Args:
            annotation (List[Union[int, float]], optional): List of pixel coordinates.
            height (Union[int, float], optional): Height used for clipping.
            width (Union[int, float], optional): Width used for clipping.
            normalized (bool, optional): Whether coordinates are normalized (0, 1) or integer pixel values. Defaults to False.

        Returns:
            List[Union[int, float]]: Clipped list of pixel coordniates.
        """
        if any(isinstance(i, list) for i in annotation):
            return [zpy.saver.Saver.clip_coordinate_list(height=height,
                                                         width=width,
                                                         normalized=normalized,
                                                         annotation=ann) for ann in annotation]
        if normalized:
            # Coordinates are in (0, 1)
            max_x, max_y = 1.0, 1.0
        else:
            # Coordinates are w.r.t image height and width
            max_x, max_y = width, height
        new_annotation = []
        # TODO: This zip unpack here is unreadable
        for x, y in zip(*[iter(annotation)]*2):
            new_x, new_y = x, y
            if x < 0:
                new_x = 0
            if y < 0:
                new_y = 0
            if x > max_x:
                new_x = max_x
            if y > max_y:
                new_y = max_y
            new_annotation.append(new_x)
            new_annotation.append(new_y)
        return new_annotation

    @staticmethod
    def clip_bbox(
        bbox: List[Union[int, float]] = None,
        height: Union[int, float] = None,
        width: Union[int, float] = None,
        normalized: bool = False,
    ) -> List[Union[int, float]]:
        """ Clip a bounding box in [x, y, width, height] format.

        Args:
            bbox (List[Union[int, float]], optional): Bounding box in [x, y, width, height] format.
            height (Union[int, float], optional): Height used for clipping.
            width (Union[int, float], optional): Width used for clipping.
            normalized (bool, optional): Whether bounding box values are normalized (0, 1) or integer pixel values. Defaults to False.

        Returns:
            List[Union[int, float]]: [description]
        """
        if normalized:
            # Coordinates are in (0, 1)
            max_x, max_y = 1.0, 1.0
        else:
            # Coordinates are w.r.t image height and width
            max_x, max_y = width, height
        new_bbox = [0] * 4
        new_bbox[0] = max(0, min(bbox[0], max_x))
        new_bbox[1] = max(0, min(bbox[1], max_y))
        new_bbox[2] = max(0, min(bbox[2], (max_x - new_bbox[0])))
        new_bbox[3] = max(0, min(bbox[3], (max_y - new_bbox[1])))
        return new_bbox



"""
    Utilities for Rendering in Blender.
"""
import logging
import os
import sys
import time
from pathlib import Path
from typing import Union, Tuple

import bpy
import gin

import zpy

log = logging.getLogger(__name__)


@gin.configurable
def make_aov_pass(
    style: str = 'instance',
) -> None:
    """ Make AOV pass in Cycles. """
    scene = zpy.blender.verify_blender_scene()
    # Make sure engine is set to Cycles
    if not (scene.render.engine == "CYCLES"):
        log.warning(' Setting render engine to CYCLES to use AOV')
        scene.render.engine = "CYCLES"
        scene.render.use_compositing = True
    # Only certain styles are available
    valid_styles = ['instance', 'category']
    assert style in valid_styles, \
        f'Invalid style {style} for AOV Output Node, must be in {valid_styles}.'
    # Go through existing passes and make sure it doesn't exist before creating
    view_layer = zpy.blender.verify_view_layer()
    if view_layer.get('aovs', None) is not None:
        for aov in view_layer['aovs']:
            if aov['name'] == style:
                log.info(f'AOV pass for {style} already exists.')
                return
    bpy.ops.cycles.add_aov()
    view_layer['aovs'][-1]['name'] = style
    view_layer.update()
    log.info(f'Created AOV pass for {style}.')


@gin.configurable
def make_aov_file_output_node(
    style: str = 'rgb',
    add_hsv: bool = True,
    add_lens_dirt: bool = False,
) -> bpy.types.CompositorNodeOutputFile:
    """ Make AOV Output nodes in Composition Graph. """
    log.info(f'Making AOV output node for {style}')

    # Only certain styles are available
    valid_styles = ['rgb', 'depth', 'instance', 'category']
    assert style in valid_styles, \
        f'Invalid style {style} for AOV Output Node, must be in {valid_styles}.'

    # Make sure scene composition is using nodes
    scene = zpy.blender.verify_blender_scene()
    if not scene.use_nodes:
        scene.use_nodes = True
    _tree = scene.node_tree

    # Get or create render layer node
    if _tree.nodes.get('Render Layers', None) is None:
        rl_node = _tree.nodes.new('CompositorNodeRLayers')
    else:
        rl_node = _tree.nodes['Render Layers']
    # assert rl_node.outputs.get(style, None) is not None, \
    #     f'Render Layer output {style} does not exist.'

    # Remove Composite Node if it exists
    composite_node = _tree.nodes.get('Composite')
    if composite_node is not None:
        _tree.nodes.remove(composite_node)

    # Instance and category require an AOV pass
    if style in ['instance', 'category']:
        zpy.render.make_aov_pass(style)

    # Visualize node shows image in workspace
    _name = f'{style} viewer'
    view_node = _tree.nodes.get(_name)
    if view_node is None:
        view_node = _tree.nodes.new('CompositorNodeViewer')
    view_node.name = _name

    # File output node renders out image
    _name = f'{style} output'
    fileout_node = _tree.nodes.get(_name)
    if fileout_node is None:
        fileout_node = _tree.nodes.new('CompositorNodeOutputFile')
    fileout_node.name = _name
    fileout_node.mute = False

    # HACK: Depth requires normalization node between layer and output
    if style == 'depth':
        # Normalization node
        _name = f'{style} normalize'
        norm_node = _tree.nodes.get(_name)
        if norm_node is None:
            norm_node = _tree.nodes.new('CompositorNodeNormalize')
        norm_node.name = _name

        # Negative inversion
        _name = f'{style} negate'
        invert_node = _tree.nodes.get(_name)
        if invert_node is None:
            invert_node = _tree.nodes.new('CompositorNodeInvert')
        invert_node.name = _name

        # Link up the nodes
        _tree.links.new(rl_node.outputs['Depth'], norm_node.inputs[0])
        _tree.links.new(norm_node.outputs[0], invert_node.inputs['Color'])
        _tree.links.new(invert_node.outputs[0], view_node.inputs['Image'])
        _tree.links.new(invert_node.outputs[0], fileout_node.inputs['Image'])
    elif style == 'rgb':
        _node = rl_node
        if add_lens_dirt:
            _node = lens_dirt_node(
                node_tree=_tree,
                input_node=rl_node
            )
            _tree.links.new(rl_node.outputs['Image'], _node.inputs['Image'])
        if add_hsv:
            _node = hsv_node(
                node_tree=_tree,
                input_node=rl_node
            )
            _tree.links.new(rl_node.outputs['Image'], _node.inputs['Image'])
        _tree.links.new(_node.outputs['Image'], view_node.inputs['Image'])
        _tree.links.new(_node.outputs['Image'], fileout_node.inputs['Image'])
    else:  # category and instance segmentation
        _tree.links.new(rl_node.outputs[style], view_node.inputs['Image'])
        _tree.links.new(rl_node.outputs[style], fileout_node.inputs['Image'])

    return fileout_node


def lens_dirt_node(
    node_tree: bpy.types.NodeTree,
    input_node: bpy.types.Node,
) -> bpy.types.Node:
    """ Add lens dirt effect to a compositor node. """
    # TODO: @kursad code to create dirt effect here.
    log.warn("NotImplemented: lens dirt ")
    return input_node


def hsv_node(
    node_tree: bpy.types.NodeTree,
    input_node: bpy.types.Node,
) -> bpy.types.Node:
    """ Adds a Hue-Saturation-Value Node."""
    hsv_node = node_tree.nodes.new('CompositorNodeHueSat')
    hsv_node.name = 'hsv'
    node_tree.links.new(input_node.outputs['Image'], hsv_node.inputs['Image'])
    return hsv_node


@gin.configurable
def render_aov(
    rgb_path: Union[str, Path] = None,
    depth_path: Union[str, Path] = None,
    iseg_path: Union[str, Path] = None,
    cseg_path: Union[str, Path] = None,
    width: int = 640,
    height: int = 480,
    hsv: Tuple[float] = None,
):
    """ Render images using AOV nodes. """
    scene = zpy.blender.verify_blender_scene()
    scene.render.resolution_x = width
    scene.render.resolution_y = height
    scene.cycles.resolution_x = width
    scene.cycles.resolution_y = height
    scene.render.resolution_percentage = 100
    scene.render.image_settings.file_format = 'PNG'

    # HACK: Prevents adding frame number to filename
    scene.frame_end = scene.frame_current
    scene.frame_start = scene.frame_current
    scene.render.use_file_extension = False
    scene.render.use_stamp_frame = False

    # Create AOV output nodes
    render_outputs = {
        'rgb': rgb_path,
        'depth': depth_path,
        'instance': iseg_path,
        'category': cseg_path,
    }
    for style, output_path in render_outputs.items():
        if output_path is not None:
            # Create output node if it is not in scene
            if not scene.use_nodes:
                scene.use_nodes = True
            output_node = scene.node_tree.nodes.get(
                f'{style} output', None)
            if output_node is None:
                output_node = make_aov_file_output_node(style=style)
            output_node.base_path = str(output_path.parent)
            output_node.file_slots[0].path = str(output_path.name)
            output_node.format.file_format = 'PNG'
            output_node.format.color_mode = 'RGB'
            if style in ['rgb']:
                output_node.format.color_depth = '8'
                output_node.format.view_settings.view_transform = 'Filmic'
                if hsv is not None:
                    hsv_node = scene.node_tree.nodes.get('hsv', None)
                    if hsv_node is not None:
                        hsv_node.inputs[1].default_value = max(
                            0, min(hsv[0], 1))
                        hsv_node.inputs[2].default_value = max(
                            0, min(hsv[1], 2))
                        hsv_node.inputs[3].default_value = max(
                            0, min(hsv[2], 2))
                    else:
                        log.warn('Render given HSV but no HSV node found.')
            if style in ['depth']:
                output_node.format.color_depth = '8'
                output_node.format.use_zbuffer = True
            if style in ['instance', 'category']:
                output_node.format.color_depth = '8'
                output_node.format.view_settings.view_transform = 'Raw'
            log.debug(
                f'Output node for {style} image pointing to {str(output_path)}')

    if render_outputs.get('rgb', None) is not None:
        # Mute segmentation and depth output nodes
        _mute_aov_file_output_node('category', mute=True)
        _mute_aov_file_output_node('instance', mute=True)
        _mute_aov_file_output_node('depth', mute=True)
        _mute_aov_file_output_node('rgb', mute=False)
        _rgb_render_settings()
        _render()

    cseg_is_on = (render_outputs.get('category', None) is not None)
    iseg_is_on = (render_outputs.get('instance', None) is not None)
    depth_is_on = (render_outputs.get('depth', None) is not None)
    if cseg_is_on or iseg_is_on or depth_is_on:
        # Un-mute segmentation and depth output nodes
        _mute_aov_file_output_node('category', mute=(not cseg_is_on))
        _mute_aov_file_output_node('instance', mute=(not iseg_is_on))
        _mute_aov_file_output_node('depth', mute=(not depth_is_on))
        _mute_aov_file_output_node('rgb', mute=True)
        _seg_render_settings()
        _render()

    # Save intermediate scene
    if log.getEffectiveLevel() == logging.DEBUG:
        # HACK: Use whatever output path is not None
        for style, output_path in render_outputs.items():
            if output_path is not None:
                break
        _filename = f'blender-debug-scene-post-aov-{output_path.stem}.blend'
        _path = output_path.parent / _filename
        zpy.blender.output_intermediate_scene(_path)

    # HACK: Rename image outputs due to stupid Blender reasons
    for style, output_path in render_outputs.items():
        if output_path is not None:
            _bad_name = str(output_path) + '%04d' % scene.frame_current
            os.rename(_bad_name, str(output_path))
            log.info(f'Rendered {style} image saved to {str(output_path)}')


def _mute_aov_file_output_node(style: str, mute: bool = True):
    """ Mute (or un-mute) an AOV output node for a style. """
    log.debug(f'Muting AOV node for {style}')
    scene = zpy.blender.verify_blender_scene()
    _output_node = scene.node_tree.nodes.get(f'{style} output', None)
    if _output_node is not None:
        _output_node.mute = mute


def _rgb_render_settings():
    """ Render settings for normal color images. """
    scene = zpy.blender.verify_blender_scene()
    scene.render.engine = "CYCLES"
    scene.render.film_transparent = False
    scene.render.dither_intensity = 1.0
    scene.render.filter_size = 1.5

    scene.cycles.samples = 128
    scene.cycles.diffuse_bounces = 4
    scene.cycles.diffuse_samples = 12

    view_layer = zpy.blender.verify_view_layer()
    view_layer.pass_alpha_threshold = 0.5

    scene.cycles.max_bounces = 4
    scene.cycles.bake_type = 'COMBINED'
    scene.cycles.use_adaptive_sampling = True
    scene.cycles.use_denoising = True
    scene.cycles.denoising_radius = 8

    scene.cycles.use_denoising = True
    scene.cycles.denoiser = 'OPENIMAGEDENOISE'

    scene.view_settings.view_transform = 'Filmic'
    # scene.sequencer_colorspace_settings.name = 'Filmic Log'

    scene.display.render_aa = '8'
    scene.display.viewport_aa = 'FXAA'
    scene.display.shading.color_type = 'TEXTURE'
    scene.display.shading.light = 'STUDIO'
    scene.display.shading.show_specular_highlight = True


def _seg_render_settings():
    """ Render settings for segmentation images. """
    scene = zpy.blender.verify_blender_scene()
    scene.render.engine = "CYCLES"
    scene.render.film_transparent = True
    scene.render.dither_intensity = 0.
    scene.render.filter_size = 0.

    scene.cycles.samples = 1
    scene.cycles.diffuse_bounces = 0
    scene.cycles.diffuse_samples = 0

    view_layer = zpy.blender.verify_view_layer()
    view_layer.pass_alpha_threshold = 0.0

    scene.cycles.max_bounces = 0
    scene.cycles.bake_type = 'EMIT'
    scene.cycles.use_adaptive_sampling = False
    scene.cycles.use_denoising = False
    scene.cycles.denoising_radius = 0

    scene.view_settings.view_transform = 'Raw'
    # scene.sequencer_colorspace_settings.name = 'Raw'

    # scene.world.use_nodes=False
    # scene.display_settings.display_device = 'None'

    scene.display.render_aa = 'OFF'
    scene.display.viewport_aa = 'OFF'
    scene.display.shading.color_type = 'MATERIAL'
    scene.display.shading.light = 'FLAT'
    scene.display.shading.show_specular_highlight = False


def _render(threads: int = 4,
            logfile: str = 'blender_render.log',
            ):
    """ Render in Blender. """
    start_time = time.time()
    scene = zpy.blender.verify_blender_scene()
    scene.render.threads = threads
    # TODO: The commented out code here only works on Linux (fails on Windows)
    # try:
    #     # HACK: This disables the blender log by redirecting output to log file
    #     # https://blender.stackexchange.com/questions/44560
    #     open(logfile, 'a').close()
    #     old = os.dup(1)
    #     sys.stdout.flush()
    #     os.close(1)
    #     os.open(logfile, os.O_WRONLY)
    # except Exception as e:
    #     log.warning(f'Render log removal raised exception {e}')
    try:
        # This is the actual render call
        bpy.ops.render.render(write_still=True)
    except Exception as e:
        log.warning(f'Render raised exception {e}')
    # try:
    #     # disable output redirection
    #     os.close(1)
    #     os.dup(old)
    #     os.close(old)
    # except Exception as e:
    #     log.warning(f'Render log removal raised exception {e}')
    duration = time.time() - start_time
    log.info(f'Rendering took {duration}s to complete.')



"""
    Gin-config utilities.
"""
import logging
from pathlib import Path
from pprint import pformat
from typing import Any, Dict, Tuple

import gin
import zpy

log = logging.getLogger(__name__)

# HACK: Conversion from zpy keys to human-readable keys
HUMAN_CONVERSION = {
    'seed': 'zpy.blender.set_seed.seed',
    'output_dir': 'zpy.saver.Saver.output_dir',
    'output_path': 'zpy.saver.Saver.output_dir',
    'num_images': 'zpy.blender.step.num_steps',
    'num_frames': 'zpy.blender.step.num_steps',
}


def replace_human_redable_kwargs(gin_bindings: Dict) -> Tuple[str, Any]:
    """ Replace any human readable versions of bindings. """
    log.info('Converting human readable bindings to gin...')
    for key, value in gin_bindings.items():
        if HUMAN_CONVERSION.get(key, None) is not None:
            log.info(f'Converted {key} to {HUMAN_CONVERSION[key]}')
            yield HUMAN_CONVERSION[key], value
        else:
            yield key, value


def parse_gin_bindings(
    gin_bindings: Dict = None,
) -> None:
    """ Parse any extra gin bindings to the config. """
    if gin_bindings is None:
        log.info(f'No additional gin bindings to parse')
    else:
        log.info(f'Parsing additional bindings: {pformat(gin_bindings)}')
        with gin.unlock_config():
            for key, value in replace_human_redable_kwargs(gin_bindings):
                try:
                    gin.bind_parameter(key, value)
                    _message = 'BOUND  '
                except:
                    _message = 'IGNORED'
                log.info(f'{_message} - {key} : {value}')


def parse_gin_config(
    gin_config: str = None,
    gin_config_dir: str = '$CONFIG',
) -> None:
    """ Parse a gin config file by path. """
    if gin_config is None:
        log.info(f'No gin file to parse.')
    else:
        gin_config_dir = zpy.files.verify_path(gin_config_dir, check_dir=True)
        if not gin_config.endswith('.gin'):
            gin_config = gin_config + '.gin'
        gin_config_filename = Path(gin_config)
        gin_config_path = gin_config_dir / gin_config_filename
        log.info(f'Parsing gin config at {gin_config_path}')
        if not gin_config_path.exists():
            raise zpy.requests.InvalidRequest(
                f'Could not find gin config at {gin_config_path}')
        gin.clear_config()
        gin.parse_config_file(str(gin_config_path))


def parse_gin_in_request(request: Dict) -> None:
    """ Parse any gin related keys in a request dict. """
    zpy.gin.parse_gin_config(gin_config=request.get('gin_config', None))
    zpy.gin.parse_gin_bindings(gin_bindings=request.get('gin_bindings', None))
    gin.finalize()



"""
    Visualization and plotting utilities.
"""
import logging
import random
import time
from pathlib import Path
from typing import Dict, List, Tuple, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.patches import Arrow, Circle, Polygon, Rectangle
from matplotlib.ticker import MaxNLocator

import zpy

log = logging.getLogger(__name__)


def pretty_axes(ax: matplotlib.axes.Axes) -> matplotlib.axes.Axes:
    """ Better looking matplotlib axes object. """
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().set_visible(False)
    ax.grid(axis='y', alpha=0.75)
    return ax


def plotter(func):
    """ Decorator for plotting images. """

    def wrapped(output_path: Union[str, Path] = None,
                show: bool = False,
                **kwargs,
                ) -> matplotlib.figure.Figure:
        output_path = zpy.files.verify_path(
            output_path, make=True, check_dir=True)
        plt.style.use('fivethirtyeight')
        plot_name, fig = func(**kwargs)
        if show:
            log.info(f'Displaying {plot_name}...')
            plt.show()
            time.sleep(1)
        output_path = output_path / (plot_name + '.png')
        plt.savefig(output_path, bbox_inches="tight", pad_inches=0)
        plt.close('all')
        return fig

    return wrapped


@plotter
def image_grid_plot(
        images: List[np.ndarray] = None,
        rows: int = 4,
        cols: int = 4,
) -> Tuple[str, matplotlib.figure.Figure]:
    """ Plots images in a grid. """
    assert images is not None, 'Images required.'
    sample_size = min(rows*cols, len(images))
    images = random.sample(images, sample_size)
    fig = plt.figure(figsize=(16, 16))
    plt.suptitle('Sample Images', fontsize=18)
    for n, image in enumerate(images):
        plt.subplot(rows, cols, n+1)
        plt.imshow(image)
        plt.grid(False)
        plt.xticks([])
        plt.yticks([])
    return 'image_grid_plot', fig


@plotter
def image_shape_plot(
    images: List[np.ndarray] = None,
) -> Tuple[str, matplotlib.figure.Figure]:
    """ Plots 2D histogram of the image shapes. """
    assert images is not None, 'Images required.'
    image_shapes = [np.shape(_) for _ in images]
    # HACK: Filter out 2D images
    image_shapes = [_ for _ in image_shapes if len(_) == 3]
    image_shape = np.asarray(image_shapes)
    fig, ax = plt.subplots(figsize=(8, 8))
    ax.hist2d(image_shape[:, 0], image_shape[:, 1], bins=(50, 50))
    ax.set(title='Histogram of Image Sizes')
    ax.set(xlabel='Width in Pixels')
    ax.set(ylabel='Height in Pixels')
    # Pixel ticks should be integers
    ax.yaxis.set_major_locator(MaxNLocator(nbins=5, integer=True))
    ax.xaxis.set_major_locator(MaxNLocator(nbins=5, integer=True))
    return 'image_shape_plot', fig


@plotter
def color_correlations_plot(
    flat_images: List[np.ndarray] = None,
) -> Tuple[str, matplotlib.figure.Figure]:
    """ Plots 2D histograms of color correlations: RG, RB, and BG. """
    assert flat_images is not None, 'Images required.'
    # HACK: Incorrect type assumption
    flat_images = flat_images[0]
    fig = plt.figure(figsize=(16, 5))
    plt.rcParams["axes.grid"] = False
    plt.suptitle('Pixel Color Correlations \n\n\n', fontsize=18)
    ax = plt.subplot(131)
    plt.hist2d(flat_images[:, 0], flat_images[:, 1],
               bins=(50, 50), density=True)
    plt.title('Red - Green', fontsize=16)
    ax = plt.subplot(132)
    plt.hist2d(flat_images[:, 0], flat_images[:, 2],
               bins=(50, 50), density=True)
    plt.title('Red - Blue', fontsize=16)
    ax = plt.subplot(133)
    plt.hist2d(flat_images[:, 1], flat_images[:, 2],
               bins=(50, 50), density=True)
    plt.title('Blue - Green', fontsize=16)
    return 'color_correlations_plot', fig


@plotter
def pixel_histograms(
    flat_images: List[np.ndarray] = None,
) -> Tuple[str, matplotlib.figure.Figure]:
    """ Plots histograms of pixel values for each color channel. """
    assert flat_images is not None, 'Images required.'
    import seaborn as sns

    # HACK: Incorrect type assumption
    flat_images = flat_images[0]
    fig = plt.figure(figsize=(16, 8))
    plt.suptitle('Pixel Histograms (Red, Green, Blue)', fontsize=18)
    ax = plt.subplot(311)
    ax = pretty_axes(ax)
    sns.histplot(flat_images[:, 0], bins=255, color='r', ax=ax)
    ax = plt.subplot(312)
    ax = pretty_axes(ax)
    sns.histplot(flat_images[:, 1], bins=255, color='g', ax=ax)
    ax = plt.subplot(313)
    ax = pretty_axes(ax)
    sns.histplot(flat_images[:, 2], bins=255, color='b', ax=ax)
    plt.tight_layout()
    return 'pixel_histograms', fig


@plotter
def category_barplot(
    categories: Dict[str, Dict] = None,
) -> Tuple[str, matplotlib.figure.Figure]:
    """ Histograms for categories and sub-categories. """
    assert categories is not None, 'categories required.'

    category_names = [c['name'] for c in categories.values()]
    category_count = [c['count'] for c in categories.values()]
    category_color = [c['color'] for c in categories.values()]
    num_categories = len(category_names)

    fig, ax = plt.subplots(figsize=(16, 6 * (num_categories + 1)))
    plt.rcParams["axes.grid"] = False

    # Category histograms
    subplot_num_rows = num_categories + 1
    subplot_num_cols = 1
    subplot_plot_idx = 1
    if num_categories >= 1:
        ax = plt.subplot(subplot_num_rows, subplot_num_cols, subplot_plot_idx)
        ax.barh(category_names, category_count, color=category_color)
        ax.set(title='Annotations per Category')
        ax.set(xlabel='Number of Annotations')
        ax.set(ylabel='Category Name')

    # Subcategory histograms
    for i, category in enumerate(categories.values()):
        subcategories = category['subcategories']
        if len(subcategories) > 1:
            ax = plt.subplot(subplot_num_rows,
                             subplot_num_cols,
                             subplot_plot_idx + i + 1)
            subcategory_count = category['subcategory_count']
            ax.barh(subcategories, category['subcategory_count'])
            ax.set(title=f'Annotations per Subcategory of {category["name"]}')
            ax.set(xlabel='Number of Annotations')
            ax.set(ylabel=f'Subcategory of {category["name"]}')
    return 'category_histograms', fig


@plotter
def draw_annotations(
    image_path: Union[str, Path] = None,
    annotations: List = None,
    categories: Dict[str, Dict] = None,
) -> None:
    """ Given an path to an image draw annotations. """
    log.info(f'draw annotations on {image_path}...')
    image = zpy.image.open_image(image_path)
    _, ax = plt.subplots()
    ax.imshow(image)
    for i, annotation in enumerate(annotations):
        log.debug(f'{i}: {annotation}')
        category_id = annotation['category_id']
        category_color = categories[category_id].get('color', None)
        if category_color is None:
            log.debug(
                'Could not find category color, using random color instead.')
            category_color = zpy.color.random_color()
        if 'num_keypoints' in annotation:
            skeleton = categories[category_id]['skeleton']
            try:
                keypoints = annotation['keypoints_xyv']
            except KeyError:
                keypoints = annotation['keypoints']
            draw_keypoints(ax, keypoints, skeleton, 'r')
        # Only draw bounding box OR segmentation
        if 'segmentation' in annotation:
            draw_segmentation(ax, annotation['segmentation'], category_color)
        if 'bbox' in annotation:
            draw_bbox(ax,
                      annotation['bbox'],
                      category_color,
                      text=annotation.get('bbox_text', None))
    plt.axis('off')
    fig = plt.gcf()
    DPI = fig.get_dpi()
    fig.set_size_inches(
        image.shape[1]/float(DPI),  # width
        image.shape[0]/float(DPI),  # height
    )
    full_name = f'{image_path.stem}_annotated'
    return full_name, fig


def draw_bbox(
        ax: matplotlib.axes.Axes,
        bbox: List,
        color: Tuple[int],
        text: str = None,
        alpha: float = 0.2,
) -> None:
    """ Draw a bounding box on the matplotlib axes object. """
    # TODO: fix the bordering in matplotlib so that the pixels
    #   line up appropriately bounding boxes are [x, y, w, h]
    log.debug(f'Drawing bbox {bbox} {color}')
    r = Rectangle((bbox[0], bbox[1]),
                  (bbox[2]),
                  (bbox[3]),
                  linewidth=3,
                  facecolor=color,
                  edgecolor=color,
                  alpha=alpha,
                  )
    # Add text above box
    if text is not None:
        ax.text(
            x=bbox[0],
            y=bbox[1],
            s=text,
            color=color,
            weight='bold',
            fontsize=6,
            ha='left',
            va='bottom',
        )
    ax.add_patch(r)


def draw_segmentation(
        ax: matplotlib.axes.Axes,
        segmentation: List,
        color: Tuple[int],
        alpha: float = 0.6,
) -> None:
    """ Draw a segmentation polygon on the matplotlib axes object. """
    log.debug(f'Drawing segmentation {segmentation} {color}')
    for seg in segmentation:
        p = Polygon(
            np.array(seg).reshape((int(len(seg)/2), 2)),
            linewidth=3,
            color=color,
            alpha=alpha)
        ax.add_patch(p)


def draw_keypoints(
        ax: matplotlib.axes.Axes,
        keypoints: List,
        skeleton: Dict,
        color: Tuple[int],
        alpha: float = 0.8,
) -> None:
    """
    Draws keypoints of an instance and follows the rules for keypoint connections
    to draw lines between appropriate keypoints.

    "keypoints": [x1,y1,v1,...,xk,yk,vk]
    - Keypoint coordinates are floats measured from the top left image corner (and are 0-indexed).
    - We recommend rounding coordinates to the nearest pixel to reduce file size.
    - v indicates visibility
            v=0: not labeled (in which case x=y=0)
            v=1: labeled but not visible
            v=2: labeled and visible

    """
    log.debug(f'Drawing keypoints')
    for k1, k2 in skeleton:

        # HACK: 0 indexed versus 1 indexed skeleton
        if min(min(skeleton)) == 1:
            k1 -= 1
            k2 -= 1

        k1_x = keypoints[3*k1 + 0]
        k1_y = keypoints[3*k1 + 1]
        k1_v = keypoints[3*k1 + 2]

        k2_x = keypoints[3*k2 + 0]
        k2_y = keypoints[3*k2 + 1]
        k2_v = keypoints[3*k2 + 2]

        if k1_v == 1:
            circle = Circle(
                (k1_x, k1_y),
                radius=5,
                edgecolor=color,
                facecolor='w',
                alpha=alpha
            )
            ax.add_patch(circle)

        if k1_v == 2:
            circle = Circle(
                (k1_x, k1_y),
                radius=5,
                edgecolor=color,
                facecolor=color,
                alpha=alpha
            )
            ax.add_patch(circle)

        if k2_v == 1:
            circle = Circle(
                (k2_x, k2_y),
                radius=5,
                edgecolor=color,
                facecolor='w',
                alpha=alpha
            )
            ax.add_patch(circle)

        if k2_v == 2:
            circle = Circle(
                (k2_x, k2_y),
                radius=5,
                edgecolor=color,
                facecolor=color,
                alpha=alpha
            )
            ax.add_patch(circle)

        if k1_v != 0 and k2_v != 0:
            line = Arrow(
                k1_x,
                k1_y,
                k2_x-k1_x,
                k2_y-k1_y,
                color=color,
                alpha=alpha
            )
            ax.add_patch(line)



from cli.utils import auth_headers, download_url
from zpy.files import to_pathlib_path
from table_logger import TableLogger
import requests
import json
import logging


log = logging.getLogger(__name__)


def create_scene(name, path, url, token):
    """ upload a local scene to ragnarok """
    endpoint = f'{url}/api/v1/scenes/'
    data = {'name': name}
    files = {'file': open(path, 'rb')}
    r = requests.post(endpoint, data=data, headers=auth_headers(token), files=files)
    if r.status_code != 201:
        log.warning(f'unable to create scene {name} from {path}')
        return
    log.info(f'created scene {name} from {path}')


def fetch_scene(name, path, url, token):
    """ fetch a scene from ragnarok """
    endpoint = f'{url}/api/v1/scenes/'
    params = {'name': name}
    r = requests.get(endpoint, params=params, headers=auth_headers(token))
    if r.status_code != 200:
        log.warning(f'Unable to fetch scenes')
        return
    response = json.loads(r.text)
    if response['count'] != 1:
        log.warning(f'Unable to find scene with name "{name}"')
        return
    scene = response['results'][0]
    endpoint = f"{url}/api/v1/scenes/{scene['id']}/download"
    r = requests.get(endpoint, headers=auth_headers(token))
    if r.status_code != 200:
        log.warning(f"Unable to get download link for scene {scene['id']}")
        return
    response = json.loads(r.text)
    name_slug = f"{scene['name'].replace(' ', '_')}-{scene['id'][:8]}.zip"
    output_path = to_pathlib_path(path) / name_slug
    download_url(response['redirect_link'], output_path)


def fetch_scenes(endpoint, token):
    """ fetch all datasets in ragnarok """
    endpoint = f'{endpoint}/api/v1/scenes/'
    r = requests.get(endpoint, headers=auth_headers(token))
    if r.status_code != 200:
        log.warning('Unable to fetch scenes')
        return
    scenes = json.loads(r.text)['results']
    tbl = TableLogger(columns='state,name,zpy_version,blender_version,created',default_colwidth=30)
    if len(scenes) == 0:
        log.info(None)
    for s in scenes:
        tbl(s['state'], s['name'], s['zpy_version'], s['blender_version'], s['created_at'])



"""
    Generic CSV dataset format.
"""
import logging
from pathlib import Path
from typing import List, Union, Callable

import gin

import zpy
from zpy.output import Output
from zpy.saver_image import ImageSaver

log = logging.getLogger(__name__)


class CSVParseError(Exception):
    """ Invalid CSV Annotation found when parsing data contents. """
    pass


@gin.configurable
class OutputCSV(Output):
    """Holds the logic for outputting CSV annotations to file."""

    ANNOTATION_FILENAME = Path('annotations.csv')

    @gin.configurable
    def output_annotations(self,
                           annotation_dict_to_csv_row_func: Callable = None,
                           annotation_path: Union[str, Path] = None,
                           header: List[str] = None,
                           ):
        """ Ouput Generic CSV annotations. """
        if annotation_dict_to_csv_row_func is None:
            raise CSVParseError(
                'Output CSV annotations requires a annotation_dict_to_csv_row_func')
        csv_data = []
        if header is not None:
            csv_data.append(header)
        for annotation in self.saver.annotations:
            row = annotation_dict_to_csv_row_func(annotation, saver=self.saver)
            if row is not None:
                csv_data.append(row)
        # Get the correct annotation path
        if annotation_path is not None:
            annotation_path = annotation_path
        elif self.saver.annotation_path is None:
            annotation_path = self.saver.output_dir / self.ANNOTATION_FILENAME
        else:
            annotation_path = self.saver.annotation_path
        # Write out annotations to file
        zpy.files.write_csv(annotation_path, csv_data)
        # Verify annotations
        parse_csv_annotations(annotation_path)


@gin.configurable
def parse_csv_annotations(
    annotation_file: Union[str, Path],
) -> None:
    """ Parse CSV annotations """
    log.info(f'Verifying CSV annotations at {annotation_file}...')
    csv_data = zpy.files.read_csv(annotation_file)
    # Make sure all the rows are the same length
    csv_data_iterable = iter(csv_data)
    try:
        length = len(next(csv_data_iterable))
    except StopIteration:
        raise CSVParseError(f'No data found in CSV at {annotation_file}')
    log.debug(f'Row length in CSV: {[length for l in csv_data_iterable]}')
    if not all(len(l) == length for l in csv_data_iterable):
        raise CSVParseError(
            f'Not all rows in the CSV have same length {length}')



"""
    Blender utilities.
"""
import inspect
import logging
import math
import random
import time
from pathlib import Path
from typing import Dict, List, Tuple, Union

import bpy
import gin
import mathutils
import numpy as np

import zpy

log = logging.getLogger(__name__)


def use_gpu(device_type='CUDA', use_cpus=True) -> None:
    """ Use GPU for rendering. """
    preferences = bpy.context.preferences
    cycles_preferences = preferences.addons["cycles"].preferences
    cuda_devices, opencl_devices = cycles_preferences.get_devices()

    if device_type == "CUDA":
        devices = cuda_devices
    elif device_type == "OPENCL":
        devices = opencl_devices
    else:
        raise RuntimeError("Unsupported device type")

    for device in devices:
        if device.type == "CPU":
            device.use = use_cpus
        else:
            device.use = True

    cycles_preferences.compute_device_type = device_type
    bpy.context.scene.cycles.device = "GPU"
    log.info(f'using devices {devices}')


@gin.configurable
def set_seed(seed: int = 0) -> None:
    """ Set the random seed. """
    log.info(f'Setting random seed to {seed}')
    if log.getEffectiveLevel() == logging.DEBUG:
        # When debugging you want to run into errors related
        # to specific permutations of the random variables, so
        # you need to vary the seed to run into them.
        seed = random.randint(1, 100)
        log.debug(f'Choosing a random random seed of {seed}')
    random.seed(seed)
    np.random.seed(seed)


@gin.configurable
def step(
    num_steps: int = 16,
    framerate: int = 0,
    start_frame: int = 1,
    refresh_ui: bool = False,
) -> int:
    """ Step logic helper for the scene. """
    assert num_steps is not None, 'Invalid num_steps'
    assert num_steps > 0, 'Invalid num_steps'
    scene = zpy.blender.verify_blender_scene()
    step_idx = 0
    if framerate > 0:
        start = scene.frame_start
        stop = scene.frame_end
        log.info(f'Animation enabled. Min frames: {start}. Max frames: {stop}')
    while step_idx < num_steps:
        log.info('-----------------------------------------')
        log.info('                   STEP                  ')
        log.info('-----------------------------------------')
        log.info(f'Simulation step {step_idx} of {num_steps}.')
        start_time = time.time()
        if framerate > 0:
            current_frame = start_frame + step_idx * framerate
            scene.frame_set(current_frame)
            log.info(f'Animation frame {scene.frame_current}')
        # # Update the step_idx for all RandomEvent and Animator instances
        # RandomEvent.step_idx = step_idx
        yield step_idx
        step_idx += 1
        duration = time.time() - start_time
        log.info(f'Simulation step took {duration}s to complete.')
        # TODO: This call is not needed in headless instances, makes loop faster
        if refresh_ui:
            refresh_blender_ui()


def connect_debugger_vscode(timeout: int = 3) -> None:
    """ Connects to a VSCode debugger.

    Based on:

    https://github.com/AlansCodeLog/blender-debugger-for-vscode

    """
    if log.getEffectiveLevel() == logging.DEBUG:
        log.debug('Starting VSCode debugger in Blender.')
        # TODO: Can we assume the user will properly set up this environment variable?
        path = '$BLENDERADDONS/blender-debugger-for-vscode/__init__.py'
        path = zpy.files.verify_path(path, make=False)
        bpy.ops.preferences.addon_install(filepath=str(path))
        bpy.ops.preferences.addon_enable(module='blender-debugger-for-vscode')
        bpy.ops.debug.connect_debugger_vscode()
        for sec in range(timeout):
            log.debug(f'You have {timeout - sec} seconds to connect!')
            time.sleep(1)


@gin.configurable
def verify_view_layer(
    view_layer_name: str = 'prod',
) -> bpy.types.ViewLayer:
    """ Get and set the view layer for a scene. """
    scene = zpy.blender.verify_blender_scene()
    view_layer = scene.view_layers.get(view_layer_name, None)
    if view_layer is None:
        log.info(f'Could not find view layer {view_layer_name}')
        # Default behavior is to use last view layer in view layer list
        view_layer = scene.view_layers[-1]
    log.info(f'Setting view layer to {view_layer.name}')
    bpy.context.window.view_layer = view_layer
    return view_layer


@gin.configurable
def verify_blender_scene(
    blender_scene_name: str = 'Prod',
) -> bpy.types.Scene:
    """ Get and set the main scene. """
    scene = bpy.data.scenes.get(blender_scene_name, None)
    if scene is None:
        log.info(f'Could not find scene {blender_scene_name}')
        # Default behavior is to use the first scene
        scene = bpy.data.scenes[0]
    log.info(f'Setting scene to {scene.name}')
    bpy.context.window.scene = scene
    return scene


def parse_config(text_name: str = 'config') -> None:
    """ Load gin config for scene """
    _text = bpy.data.texts.get(text_name, None)
    if _text is None:
        log.warning(f'Could not find {text_name} in texts.')
        return
    log.info(f'Loading gin config {text_name}')
    gin.enter_interactive_mode()
    with gin.unlock_config():
        gin.parse_config(_text.as_string())
        gin.finalize()


def run_text(text_name: str = 'run') -> None:
    """ Run a text script in Blender. """
    _text = bpy.data.texts.get(text_name, None)
    if _text is None:
        log.warning(f'Could not find {text_name} in texts.')
        return
    _ctx = bpy.context.copy()
    _ctx['edit_text'] = _text
    bpy.ops.text.run_script(_ctx)


def connect_addon(name: str = 'zpy_addon') -> None:
    """ Connects a Blender AddOn. """
    log.debug(f'Connecting Addon {name}.')
    path = f'$BLENDERADDONS/{name}/__init__.py'
    path = zpy.files.verify_path(path, make=False)
    bpy.ops.preferences.addon_install(filepath=str(path))
    bpy.ops.preferences.addon_enable(module=name)


def output_intermediate_scene(path: Union[str, Path] = None) -> None:
    """ Output intermediate saved scene. """
    if path is None:
        path = zpy.files.default_temp_path() / 'blender-debug-scene-tmp.blend'
    path = zpy.files.verify_path(path, make=False)
    log.debug(f'Saving intermediate scene to {path}')
    bpy.ops.wm.save_as_mainfile(filepath=str(path), compress=False, copy=True)


def refresh_blender_ui() -> None:
    """ Refresh blender in the middle of a script.

    Does not work on headless instances.
    """
    log.debug(f'Refreshing Blender UI.')
    bpy.ops.wm.redraw_timer(type='DRAW_WIN_SWAP', iterations=1)
    view_layer = zpy.blender.verify_view_layer()
    view_layer.update()


def load_scene(
    path: Union[str, Path],
    auto_execute_scripts: bool = True,
) -> None:
    """ Load a scene from a *.blend file. """
    # HACK: Clear out scene of cameras and lights
    clear_scene(['CAMERA', 'LIGHT'])
    path = zpy.files.verify_path(path, make=False)
    log.debug(f'Loading scene from {str(path)}.')
    with bpy.data.libraries.load(str(path)) as (data_from, data_to):
        for attr in dir(data_to):
            setattr(data_to, attr, getattr(data_from, attr))
    # HACK: Delete current empty scene
    bpy.ops.scene.delete()
    # HACK: Delete extra workspaces that are created e.g. 'Animation.001'
    _workspaces = [ws for ws in bpy.data.workspaces if '.0' in ws.name]
    bpy.data.batch_remove(ids=_workspaces)
    # Allow execution of scripts inside loaded scene
    if auto_execute_scripts:
        log.warning('Allowing .blend file to run scripts automatically')
        log.warning(
            '   this is unsafe for blend files from an untrusted source')
        bpy.context.preferences.filepaths.use_scripts_auto_execute = auto_execute_scripts


def clear_scene(to_clear: List = ["MESH"]) -> None:
    """ Empty out the scene. """
    log.debug('Deleting all mesh objects in the scene.')
    for obj in bpy.data.objects:
        if obj.type in to_clear:
            bpy.data.objects.remove(obj)


def load_text_from_file(
    path: Union[str, Path],
    text_name: str = '',
) -> None:
    """ Load a file into Blender's internal text UI. """
    path = zpy.files.verify_path(path)
    if bpy.data.texts.get(text_name, None) is None:
        _text = bpy.data.texts.load(str(path), internal=True)
        _text.name = text_name
    else:
        bpy.data.texts[text_name].from_string(path.read_text())


def scene_information() -> Dict:
    """ Get the run() function kwargs. """
    log.info(f'Collecting scene information')
    run_script = bpy.data.texts.get('run', None)
    if run_script is None:
        raise ValueError('No run script found in scene.')
    # HACK: Gin is confused by the as_module() call
    gin.enter_interactive_mode()
    run_script_module = bpy.data.texts['run'].as_module()
    scene_doc = inspect.getdoc(run_script_module)

    run_function = None
    for name, value in inspect.getmembers(run_script_module):
        if name == 'run':
            run_function = value
    if run_function is None:
        raise ValueError('No run() function found in run script.')
    if not inspect.isfunction(run_function):
        raise ValueError('run() is not a function in run script.')

    run_kwargs = []
    for param in inspect.signature(run_function).parameters.values():
        _kwarg = {}
        _kwarg['name'] = param.name
        _kwarg['type'] = str(param.annotation)
        _kwarg['default'] = param.default
        run_kwargs.append(_kwarg)

    scene = zpy.blender.verify_blender_scene()
    _ = {
        'name': scene.zpy_scene_name,
        'version': scene.zpy_scene_version,
        'description': scene_doc,
        'run_kwargs': run_kwargs,
        'export_date':  time.strftime("%m%d%Y_%H%M_%S"),
        'zpy_version': zpy.__version__,
        'zpy_path': zpy.__file__,
        'blender_version': '.'.join([str(_) for _ in bpy.app.version]),
    }
    log.info(f'{_}')
    return _



"""
    Utilities for Materials in Blender.
"""
import logging
import random
from pathlib import Path
from typing import Tuple, Union

import bpy
import gin
import numpy as np

import zpy

log = logging.getLogger(__name__)


def verify(mat: Union[str, bpy.types.Material], check_none=True) -> bpy.types.Material:
    """ Return material given name or Material type object. """
    if isinstance(mat, str):
        mat = bpy.data.materials.get(mat)
    if check_none and mat is None:
        raise ValueError(f'Could not find material {mat}.')
    return mat


_SAVED_MATERIALS = {}


def save_material_props(mat: Union[str, bpy.types.Material]) -> None:
    """ Save a pose (rot and pos) to dict. """
    log.info(f'Saving material properties for {mat.name}')
    _SAVED_MATERIALS[mat.name] = get_mat_props(mat)


def restore_material_props(mat: Union[str, bpy.types.Material]) -> None:
    """ Restore an object to a position. """
    log.info(f'Restoring material properties for {mat.name}')
    set_mat_props(mat, _SAVED_MATERIALS[mat.name])


def restore_all_material_props() -> None:
    """ Restore all jittered materials to original look. """
    for mat_name, mat_props in _SAVED_MATERIALS.items():
        set_mat_props(mat_name, mat_props)


def get_mat_props(
    mat: Union[str, bpy.types.Material],
) -> Tuple[float]:
    """ Get (some of the) material properties. """
    mat = verify(mat)
    bsdf_node = mat.node_tree.nodes.get('Principled BSDF')
    if bsdf_node is None:
        log.warning(f'No BSDF node in {mat.name}')
        return (0.0, 0.0, 0.0)
    return (
        bsdf_node.inputs['Roughness'].default_value,
        bsdf_node.inputs['Metallic'].default_value,
        bsdf_node.inputs['Specular'].default_value,
    )


def set_mat_props(
    mat: Union[str, bpy.types.Material],
    prop_tuple: Tuple[float]
) -> None:
    """ Set (some of the) material properties. """
    mat = verify(mat)
    # TODO: Work backwards from Material output node instead of
    #       assuming a 'Principled BSDF' node
    bsdf_node = mat.node_tree.nodes.get('Principled BSDF', None)
    if bsdf_node is None:
        log.warning(f'No BSDF node in {mat.name}')
        return
    bsdf_node.inputs['Roughness'].default_value += prop_tuple[0]
    bsdf_node.inputs['Metallic'].default_value += prop_tuple[1]
    bsdf_node.inputs['Specular'].default_value += prop_tuple[2]


@gin.configurable
def jitter(
    mat: Union[str, bpy.types.Material],
    std: float = 0.2,
    save_first_time: bool = True,
) -> None:
    """ Randomize an existing material a little. """
    mat = verify(mat)
    # Save the material props first time jitter is called
    # and restore before jittering every subsequent time
    if save_first_time:
        if _SAVED_MATERIALS.get(mat.name, None) is None:
            save_material_props(mat)
        else:
            restore_material_props(mat)
    log.info(f'Jittering material {mat.name}')
    mat_props = get_mat_props(mat)
    jittered_mat_props = tuple(
        map(lambda p: p + random.gauss(0, std), mat_props))
    set_mat_props(mat, jittered_mat_props)


@gin.configurable
def make_mat_from_texture(
    texture_path: Union[str, Path],
    name: str = None,
) -> bpy.types.Material:
    """ Makes a material from a texture or color."""
    texture_path = zpy.files.verify_path(texture_path, make=False)
    if name is None:
        name = texture_path.stem
    mat = bpy.data.materials.get(name, None)
    if mat is None:
        log.debug(f'Material {name} does not exist, creating it.')
        mat = bpy.data.materials.new(name=name)
    mat.use_nodes = True
    bsdf_node = mat.node_tree.nodes.get('Principled BSDF')
    out_node = mat.node_tree.nodes.get('Material Output')
    tex_node = mat.node_tree.nodes.new('ShaderNodeTexImage')
    tex_node.name = 'ImageTexture'
    coord_node = mat.node_tree.nodes.new('ShaderNodeTexCoord')
    bpy.ops.image.open(filepath=str(texture_path))
    tex_node.image = bpy.data.images[texture_path.name]
    tex_node.image.colorspace_settings.name = 'Filmic Log'
    mat.node_tree.links.new(tex_node.outputs[0], bsdf_node.inputs[0])
    mat.node_tree.links.new(coord_node.outputs[0], tex_node.inputs[0])
    tex_node.image.reload()
    return mat


@gin.configurable
def make_mat_from_color(
    color: Tuple[float],
    name: str = None,
) -> bpy.types.Material:
    """ Makes a material from a texture or color."""
    if name is None:
        name = str(color)
    mat = bpy.data.materials.get(name, None)
    if mat is None:
        log.debug(f'Material {name} does not exist, creating it.')
        mat = bpy.data.materials.new(name=name)
    mat.use_nodes = True
    bsdf_node = mat.node_tree.nodes.get('Principled BSDF')
    out_node = mat.node_tree.nodes.get('Material Output')
    mat.node_tree.nodes.remove(bsdf_node)
    bsdf_node = mat.node_tree.nodes.new('ShaderNodeBsdfDiffuse')
    bsdf_node.inputs['Color'].default_value = color + (1.,)
    mat.node_tree.links.new(out_node.inputs[0], bsdf_node.outputs[0])
    return mat


def set_mat(
    obj: Union[str, bpy.types.Object],
    mat: Union[str, bpy.types.Material],
    recursive: bool = True,
) -> None:
    """ Recursively sets object material.

    Allows string material and object names as input.
    """
    obj = zpy.objects.verify(obj)
    mat = zpy.material.verify(mat)
    if hasattr(obj, 'active_material'):
        log.debug(f'Setting object {obj.name} material {mat.name}')
        obj.active_material = mat
    else:
        log.warning('Object does not have material property')
        return
    # Recursively change material on all children of object
    if recursive:
        for child in obj.children:
            set_mat(child, mat)


@gin.configurable
def make_aov_material_output_node(
    mat: bpy.types.Material = None,
    obj: bpy.types.Object = None,
    style: str = 'instance',
) -> None:
    """ Make AOV Output nodes in Composition Graph. """
    scene = zpy.blender.verify_blender_scene()
    # Make sure engine is set to Cycles
    if not (scene.render.engine == "CYCLES"):
        log.warning(' Setting render engine to CYCLES to use AOV')
        scene.render.engine == "CYCLES"
    # Only certain styles are available
    valid_styles = ['instance', 'category']
    assert style in valid_styles, \
        f'Invalid style {style} for AOV material output node, must be in {valid_styles}.'

    # HACK: multiple material slots
    all_mats = []

    # Use material
    if mat is not None:
        all_mats = [mat]
    # Get material from object
    elif obj is not None:
        if obj.active_material is None:
            log.debug(f'No active material found for {obj.name}')
            return
        if len(obj.material_slots) > 1:
            for mat in obj.material_slots:
                all_mats.append(mat.material)
        else:
            all_mats.append(obj.active_material)
    else:
        raise ValueError('Must pass in an Object or Material')

    # HACK: multiple material slots
    for mat in all_mats:

        # Make sure material is using nodes
        if not mat.use_nodes:
            mat.use_nodes = True
        _tree = mat.node_tree

        # Vertex Color Node
        _name = f'{style} Vertex Color'
        vertexcolor_node = _tree.nodes.get(_name)
        if vertexcolor_node is None:
            vertexcolor_node = _tree.nodes.new('ShaderNodeVertexColor')
        vertexcolor_node.layer_name = style
        vertexcolor_node.name = _name

        # AOV Output Node
        _name = style
        # HACK: property "name" of ShaderNodeOutputAOV behaves strangely with .get()
        aovoutput_node = None
        for _node in _tree.nodes:
            if _node.name == _name:
                aovoutput_node = _node
        if aovoutput_node is None:
            aovoutput_node = _tree.nodes.new('ShaderNodeOutputAOV')
        aovoutput_node.name = style
        _tree.links.new(vertexcolor_node.outputs['Color'],
                        aovoutput_node.inputs['Color'])



#!/usr/bin/env python
import os
from distutils.core import setup
from setuptools import find_packages
import versioneer


def get_requirements_from_file(filepath):
    requires = []
    with open(filepath, 'r') as f:
        requires.append(f.readline())
    return requires


def get_long_description(filepath):
    with open(filepath, 'r', encoding='utf-8') as fh:
        long_description = fh.read()
    return long_description


setup(
    name='zpy-zumo',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Create synthetic data with Blender.',
    long_description=get_long_description('README.md'),
    long_description_content_type="text/markdown",
    author='Zumo Labs',
    author_email='infra@zumolabs.ai',
    url='https://github.com/ZumoLabs/zpy',
    packages=find_packages(),
    install_requires=get_requirements_from_file('requirements.txt'),
    include_package_data=True,
    entry_points='[console_scripts]\nzpy=cli.cli:cli\n',
)



"""
    Object utilities for Blender Python.
"""
import logging
import random
from pathlib import Path
from typing import Any, List, Tuple, Union

import bpy
import gin
import mathutils
import numpy as np

import zpy

log = logging.getLogger(__name__)


def verify(obj: Union[str, bpy.types.Object], check_none=True) -> bpy.types.Object:
    """ Return object given name or Object type object. """
    if isinstance(obj, str):
        obj = bpy.data.objects.get(obj)
    if check_none and obj is None:
        raise ValueError(f'Could not find object {obj}.')
    return obj


def load_blend_obj(
        name: str,
        path: Union[str, Path],
        link: bool = False,
) -> bpy.types.Object:
    """ Load object from blend file. """
    path = zpy.files.verify_path(path, make=False)
    scene = zpy.blender.verify_blender_scene()
    with bpy.data.libraries.load(str(path), link=link) as (data_from, data_to):
        for from_obj in data_from.objects:
            if from_obj.startswith(name):
                log.debug(f'Loading obj {from_obj} from {str(path)}.')
                data_to.objects.append(from_obj)
    # Copy objects over to the current scene
    for obj in data_to.objects:
        scene.collection.objects.link(obj)
    for texture_folder_name in ['Textures', 'textures','TEX']:
        texture_dir = path.parent / texture_folder_name
        if texture_dir.exists():
            bpy.ops.file.find_missing_files(directory=str(texture_dir))
            break
    return bpy.data.objects[name]


def select(obj: Union[bpy.types.Object, str]) -> None:
    """ Select an object. """
    obj = verify(obj)
    view_layer = zpy.blender.verify_view_layer()
    # TODO: This sometimes does not work due to context issues
    log.debug(f'Selecting obj: {obj.name}')
    bpy.ops.object.select_all(action='DESELECT')
    view_layer.objects.active = obj
    bpy.data.objects[obj.name].select_set(True, view_layer=view_layer)


def delete_obj(obj: Union[bpy.types.Object, str]) -> None:
    """ Delete an object. """
    obj = verify(obj)
    select(obj)
    log.debug(f'Removing obj: {obj.name}')
    bpy.ops.object.delete()
    # bpy.data.objects.remove(obj, do_unlink=True)


def is_inside(
    point: mathutils.Vector,
    obj: bpy.types.Object,
) -> bool:
    """ Is point inside a mesh.

    From:

    https://blender.stackexchange.com/questions/31693/how-to-find-if-a-point-is-inside-a-mesh

    """
    is_found, closest_point, normal, _ = obj.closest_point_on_mesh(point)
    if not is_found:
        return False
    p2 = closest_point - point
    v = p2.dot(normal)
    return not(v < 0.0)


def for_obj_in_selected_objs(context) -> bpy.types.Object:
    """ Safe iterable for selected objects. """
    zpy.blender.verify_view_layer()
    for obj in context.selected_objects:
        # Only meshes or empty objects TODO: Why the empty objects
        if not (obj.type == 'MESH' or obj.type == 'EMPTY'):
            continue
        # Make sure object exists in the scene
        if bpy.data.objects.get(obj.name, None) is None:
            continue
        context.view_layer.objects.active = obj
        yield obj


def for_obj_in_collections(
    collections: List[bpy.types.Collection],
) -> bpy.types.Object:
    """ Yield objects in list of collection. """
    for collection in collections:
        for obj in collection.all_objects:
            if obj.type == 'MESH':
                # This gives you direct access to data block
                yield bpy.data.objects[obj.name]


def toggle_hidden(
    obj: Union[bpy.types.Object, str],
    hidden: bool = True,
    filter_string: str = None,
) -> None:
    """ Recursive function to make object and children invisible.

    Optionally filter by a string in object name.

    """
    obj = verify(obj)
    if hasattr(obj, 'hide_render') and hasattr(obj, 'hide_viewport'):
        if (filter_string is None) or (filter_string in obj.name):
            log.debug(f'Hiding object {obj.name}')
            bpy.data.objects[obj.name].select_set(True)
            bpy.data.objects[obj.name].hide_render = hidden
            bpy.data.objects[obj.name].hide_viewport = hidden
        else:
            log.debug(
                f'Object {obj.name} does not contain filter string {filter_string}')
    else:
        log.warning('Object does not have hide properties')
        return
    for child in obj.children:
        toggle_hidden(child, hidden=hidden, filter_string=filter_string)


def randomly_hide_within_collection(
    collections: List[bpy.types.Collection],
    chance_to_hide: float = 0.9,
) -> None:
    """ Randomly hide objects in a list of collections. """
    to_hide = []
    for obj in for_obj_in_collections(collections):
        if random.random() < chance_to_hide:
            to_hide.append(obj.name)
    # HACK: hide objects by name, this causes segfault
    # if done in the for loop above, due to some kind of
    # pass by reference vs by value shenaniganry going on
    # with blender python sitting on top of blender C
    for name in to_hide:
        bpy.data.objects[name].select_set(True)
        bpy.data.objects[name].hide_render = True
        bpy.data.objects[name].hide_viewport = True


def segment(
    obj: Union[bpy.types.Object, str],
    name: str = 'default',
    color: Tuple[float] = None,
    as_category: bool = False,
    as_single: bool = False,
) -> None:
    """ Segment an object."""
    obj = verify(obj)
    if color is None:
        color = zpy.color.random_color(output_style='frgb')
    obj.color = zpy.color.frgb_to_frgba(color)
    if as_category:
        obj.seg.category_name = name
        obj.seg.category_color = color
        seg_type = 'category'
    else:
        obj.seg.instance_name = name
        obj.seg.instance_color = color
        seg_type = 'instance'
    # Make sure object material is set up correctly with AOV nodes
    populate_vertex_colors(obj, zpy.color.frgb_to_frgba(color), seg_type)
    zpy.material.make_aov_material_output_node(obj=obj, style=seg_type)
    # Recursively add property to children objects
    if as_single:
        for child in obj.children:
            segment(
                obj=child,
                name=name,
                color=color,
                as_category=as_category,
                as_single=as_single,
            )


def populate_vertex_colors(
        obj: Union[bpy.types.Object, str],
        color_rgba: Tuple[float],
        seg_type: str = 'instance',
) -> None:
    """Fill the given Vertex Color Layer with the color parameter values"""
    obj = verify(obj)
    if not obj.type == 'MESH':
        log.warning(f'Object {obj.name} is not a mesh, has no vertices.')
        return
    # TODO: Is this select needed?
    # select(obj)
    # Remove any existing vertex color data
    if len(obj.data.sculpt_vertex_colors):
        for vcol in obj.data.sculpt_vertex_colors.keys():
            if seg_type in vcol:
                obj.data.sculpt_vertex_colors.remove(
                    obj.data.sculpt_vertex_colors[seg_type])
    # Remove any existing vertex color data
    if len(obj.data.vertex_colors):
        for vcol in obj.data.vertex_colors.keys():
            if seg_type in vcol:
                obj.data.vertex_colors.remove(
                    obj.data.vertex_colors[seg_type])
    # Add new vertex color data
    obj.data.vertex_colors.new(name=seg_type)
    obj.data.sculpt_vertex_colors.new(name=seg_type)
    # Iterate through each vertex in the mesh
    for i, _ in enumerate(obj.data.vertices):
        obj.data.sculpt_vertex_colors[seg_type].data[i].color = color_rgba


def random_position_within_constraints(
    obj: Union[bpy.types.Object, str],
) -> None:
    """ Randomize position of object within constraints. """
    obj = verify(obj)
    # Make sure object has constraints
    _constraints = obj.constraints.get('Limit Location', None)
    if _constraints is not None:
        obj.location.x = random.uniform(
            obj.constraints['Limit Location'].min_x,
            obj.constraints['Limit Location'].max_x,
        )
        obj.location.y = random.uniform(
            obj.constraints['Limit Location'].min_y,
            obj.constraints['Limit Location'].max_y,
        )
        obj.location.z = random.uniform(
            obj.constraints['Limit Location'].min_z,
            obj.constraints['Limit Location'].max_z,
        )


@gin.configurable
def copy(
    obj: Union[bpy.types.Object, str],
    name: str = None,
    collection: bpy.types.Collection = None,
    is_library_object: bool = False,
) -> bpy.types.Object:
    """ Create a copy of the object. """
    obj = verify(obj)
    new_obj = bpy.data.objects.new(obj.name, obj.data)
    if name is not None:
        new_obj.name = name
    if collection is not None:
        collection.objects.link(new_obj)
    else:
        # Add to scene collection if no collection given
        scene = zpy.blender.verify_blender_scene()
        scene.collection.objects.link(obj)
    # TODO: Library Overriding functions
    if is_library_object:
        log.warning(
            f'Making mesh and material data local for obj {new_obj.name}')
        new_obj.data.make_local()
        for i in range(len(new_obj.material_slots)):
            bpy.data.objects[new_obj.name].material_slots[i].material.make_local()
        # Original object reference is lost if local copies are made
        new_obj = bpy.data.objects[new_obj.name]
    return new_obj


def translate(
    obj: Union[bpy.types.Object, str],
    translation: Tuple[float] = (0, 0, 0),
) -> None:
    """ Translate an object (in blender units). """
    obj = verify(obj)
    log.debug(f'Translating object {obj.name} by {translation}')
    log.debug(f'Before - obj.matrix_world\n{obj.matrix_world}')
    select(obj)
    bpy.ops.transform.translate(value=translation)
    log.debug(f'After - obj.matrix_world\n{obj.matrix_world}')


def rotate(
    obj: Union[bpy.types.Object, str],
    rotation: float = 0,
    axis: str = 'Z'
) -> None:
    """ Rotate an object (in radians) """
    obj = verify(obj)
    log.info(f'Rotating object {obj.name} by {rotation} radians around {axis} axis. ')
    log.debug(f'Before - obj.matrix_world\n{obj.matrix_world}')
    select(obj)
    bpy.ops.transform.rotate(value=rotation, orient_axis=axis)
    log.debug(f'After - obj.matrix_world\n{obj.matrix_world}')


def scale(
    obj: Union[bpy.types.Object, str],
    scale: Tuple[float] = (1.0, 1.0, 1.0)
) -> None:
    """ Scale an object """
    obj = verify(obj)
    log.info(f'Scaling object {obj.name} by {scale}')
    log.debug(f'Before - obj.matrix_world\n{obj.matrix_world}')
    select(obj)
    bpy.ops.transform.resize(value=scale)
    log.debug(f'After - obj.matrix_world\n{obj.matrix_world}')


def jitter(
    obj: Union[bpy.types.Object, str],
    translate_range: Tuple[Tuple[float]] = (
        (-0.05, 0.05),
        (-0.05, 0.05),
        (-0.05, 0.05),
    ),
    rotate_range: Tuple[Tuple[float]] = (
        (-0.05, 0.05),
        (-0.05, 0.05),
        (-0.05, 0.05),
    ),
    scale_range: Tuple[Tuple[float]] = (
        (1.0, 1.0),
        (1.0, 1.0),
        (1.0, 1.0),
    ),
) -> None:
    """ Apply random scale (blender units) and rotation (radians) to object """
    obj = verify(obj)
    translate(obj,
              translation=(
                  random.uniform(translate_range[0][0], translate_range[0][1]),
                  random.uniform(translate_range[1][0], translate_range[1][1]),
                  random.uniform(translate_range[2][0], translate_range[2][1]),
              ))
    rotate(obj,
           rotation=random.uniform(rotate_range[0][0], rotate_range[0][1]),
           axis='X',
           )
    rotate(obj,
           rotation=random.uniform(rotate_range[1][0], rotate_range[1][1]),
           axis='Y',
           )
    rotate(obj,
           rotation=random.uniform(rotate_range[2][0], rotate_range[2][1]),
           axis='Z',
           )
    scale(obj,
          scale=(
              random.uniform(scale_range[0][0], scale_range[0][1]),
              random.uniform(scale_range[1][0], scale_range[1][1]),
              random.uniform(scale_range[2][0], scale_range[2][1]),
          ))
    zpy.blender.verify_view_layer()
    bpy.context.view_layer.update()


_SAVED_POSES = {}


def save_pose(
    obj: Union[bpy.types.Object, str],
    pose_name: str,
) -> None:
    """ Save a pose (rot and pos) to dict. """
    obj = verify(obj)
    log.info(f'Saving pose {pose_name} based on object {obj.name}')
    _SAVED_POSES[pose_name] = obj.matrix_world.copy()


def restore_pose(
    obj: Union[bpy.types.Object, str],
    pose_name: str,
) -> None:
    """ Restore an object to a position. """
    obj = verify(obj)
    log.info(f'Restoring pose {pose_name} to object {obj.name}')
    obj.matrix_world = _SAVED_POSES[pose_name]



import os
import yaml
import logging
from requests.auth import HTTPBasicAuth
from zpy.files import to_pathlib_path
import requests

log = logging.getLogger(__name__)

ENDPOINTS = {
    'local': 'http://localhost:8000',
    'stage': 'https://ragnarok.stage.zumok8s.org',
    'prod':  'https://ragnarok.zumok8s.org'
}
CONFIG_FILE = '~/.zpy/config.yaml'


def initialize_config():
    """ Initialize zpy config if missing """
    path = to_pathlib_path(os.path.expanduser(CONFIG_FILE))
    if path.exists():
        log.debug(f'found zpy config {path}')
        config = read_config()
        log.debug(config)
        return
    CONFIG = {
        'ENVIRONMENT': 'prod',
        'TOKEN': None,
        'ENDPOINT': ENDPOINTS['prod']
    }
    log.debug(f'initializing new zpy config {path}...')
    path.parent.mkdir(parents=True, exist_ok=True)
    log.debug(CONFIG)
    write_config(CONFIG)
    log.info('please login to fetch token')


def login(username, password):
    config = read_config()
    endpoint = f"{config['ENDPOINT']}/auth/login/"
    log.info(f'login {endpoint}')
    r = requests.post(endpoint, auth=HTTPBasicAuth(username, password))
    if r.status_code != 200:
        log.warning(f'unable to login user {username}')
        return
    token = r.json()['token']
    write_token(token)
    log.warning(f'successful login for user {username}')


def write_token(token):
    config = read_config()
    config['TOKEN'] = token
    write_config(config)


def switch_env(env):
    config = read_config()
    if config['ENVIRONMENT'] == env:
        return
    config['ENVIRONMENT'] = env
    config['ENDPOINT'] = ENDPOINTS[env]
    config['TOKEN'] = None
    write_config(config)
    log.info('please login to fetch token')


def read_config():
    path = to_pathlib_path(os.path.expanduser(CONFIG_FILE))
    log.debug(f'Reading zpy config {path}')
    with path.open() as f:
        config = yaml.load(f, Loader=yaml.FullLoader)
    return config


def write_config(config):
    path = to_pathlib_path(os.path.expanduser(CONFIG_FILE))
    log.debug(f'Writing zpy config {path}')
    with path.open('w') as f:
        yaml.dump(config, f)



"""
    Rendering panel and functions.
"""
import importlib
import logging
from pathlib import Path

import bpy
import zpy
from bpy.types import Operator

log = logging.getLogger(__name__)

if "bpy" in locals():
    importlib.reload(zpy)


class RunOperator(Operator):
    """ Launch the run script in Blender's texts. """
    bl_idname = "scene.zpy_run"
    bl_label = "Run"
    bl_description = "Launch the run script in Blender's texts."
    bl_category = "ZPY"
    bl_options = {'REGISTER'}

    def execute(self, context):
        # Save the state of the scene before the run script was executed
        bpy.ops.wm.save_mainfile()
        try:
            zpy.blender.use_gpu()
            zpy.blender.parse_config('config')
            zpy.blender.run_text('run')
        except Exception as e:
            log.error(f'Executing script failed with exception {e}')
        # Return to the state of the scene before the run script was executed
        bpy.ops.wm.revert_mainfile()
        return {'FINISHED'}


class RenderOperator(Operator):
    """ Render out single image (rgb, segmented, depth). """
    bl_idname = "scene.zpy_render"
    bl_label = "Render"
    bl_description = "Render out segmented images."
    bl_category = "ZPY"
    bl_options = {'REGISTER'}

    @classmethod
    def poll(cls, context):
        # TODO: Make sure scene is good to render?
        return True

    def execute(self, context):
        context.space_data.shading.color_type = 'OBJECT'

        # Image names
        rgb_image_name = zpy.files.make_rgb_image_name(0)
        cseg_image_name = zpy.files.make_cseg_image_name(0)
        iseg_image_name = zpy.files.make_iseg_image_name(0)
        depth_image_name = zpy.files.make_depth_image_name(0)

        # Output path
        output_path = Path(context.scene.zpy_output_path)

        # Save renders to file
        zpy.render.render_aov(
            rgb_path=output_path / rgb_image_name,
            iseg_path=output_path / iseg_image_name,
            cseg_path=output_path / cseg_image_name,
            depth_path=output_path / depth_image_name,
            width=640,
            height=480,
        )

        return {'FINISHED'}


class SCENE_PT_ExecutePanel(bpy.types.Panel):
    """ UI for the addon that is visible in Blender. """
    bl_idname="SCENE_PT_ExecutePanel"
    bl_space_type = "VIEW_3D"
    bl_region_type = "UI"
    bl_label = "Execute"
    bl_category = "ZPY"

    def draw(self, context):
        layout = self.layout
        scene = context.scene

        row = layout.row()
        row.operator(
            'scene.zpy_render',
            text='Render',
            icon='FILE_IMAGE',
        )
        row = layout.row()
        row.operator(
            'scene.zpy_run',
            text='Run',
            icon='TRIA_RIGHT',
        )


"""
    Camera utilities.
"""
import logging
import math
from pathlib import Path
from typing import Dict, List, Tuple, Union

import bpy
import bpy_extras
import gin
import mathutils

import zpy

log = logging.getLogger(__name__)


def look_at(
    obj: Union[bpy.types.Object, str],
    target: Union[Tuple[float], mathutils.Vector],
    roll: float = 0,
) -> None:
    """
    Rotate obj to look at target

    :arg obj: the object to be rotated. Usually the camera
    :arg target: the location (3-tuple or Vector) to be looked at
    :arg roll: The angle of rotation about the axis from obj to target in degrees.

    Based on: https://blender.stackexchange.com/a/5220/12947 (ideasman42)
    """
    obj = zpy.objects.verify(obj)
    if not isinstance(target, mathutils.Vector):
        target = mathutils.Vector(target)
    loc = obj.location
    # direction points from the object to the target
    direction = target - obj.location
    quat = direction.to_track_quat('-Z', 'Y')
    # /usr/share/blender/scripts/addons/add_advanced_objects_menu/arrange_on_curve.py
    quat = quat.to_matrix().to_4x4()
    # convert roll from radians to degrees
    roll_matrix = mathutils.Matrix.Rotation(math.radians(roll), 4, 'Z')
    # remember the current location, since assigning to obj.matrix_world changes it
    loc = loc.to_tuple()
    obj.matrix_world = quat @ roll_matrix
    obj.location = loc


@gin.configurable
def camera_xyz(
    loc: mathutils.Vector,
    camera: bpy.types.Object = None,
    fisheye_lens: bool = False,
) -> Tuple[float]:
    """ Get camera image xy coordinates of point in scene.

    - (0, 0) is the bottom left of the camera frame.
    - (1, 1) is the top right of the camera frame.
    - Values outside 0-1 are also supported.
    - A negative ‘z’ value means the point is behind the camera.

    """
    scene = zpy.blender.verify_blender_scene()
    if camera is None:
        camera = scene.camera
    point = bpy_extras.object_utils.world_to_camera_view(scene, camera, loc)
    if point[2] < 0:
        log.debug('Point is behind camera')

    # Fix the point based on camera distortion
    if fisheye_lens:
        log.debug('Correcting for fisheye distortion')

        # HACK: There should be a better place to put this
        bpy.data.cameras[0].lens_unit = 'FOV'
        bpy.data.cameras[0].lens = 18.

        # Based on https://blender.stackexchange.com/questions/40702/how-can-i-get-the-projection-matrix-of-a-panoramic-camera-with-a-fisheye-equisol?noredirect=1&lq=1
        # Note this assumes 180 degree FOV
        cam = bpy.data.cameras[camera.name]
        f = cam.lens
        w = cam.sensor_width
        h = cam.sensor_height

        p = camera.matrix_world.inverted() @ loc
        p.normalize()

        # Calculate our angles
        phi = math.atan2(p.y, p.x)
        l = (p.x**2 + p.y**2)**(1/2)
        theta = math.asin(l)

        # Equisolid projection
        r = 2.0 * f * math.sin(theta / 2)

        u = r * math.cos(phi) / w + 0.5
        v = r * math.sin(phi) / h + 0.5

        # x = u * scene.render.resolution_x
        # y = v * scene.render.resolution_y
        # TODO: The value of point[2] here is not exactly correct ...
        return u, v, point[2]

    else:
        return point[0], point[1], point[2]


def is_child_hit(obj: bpy.types.Object, hit_obj: bpy.types.Object) -> bool:
    """ Recursive function to check if child is the hit object. """
    if obj == hit_obj:
        return True
    else:
        for child in obj.children:
            if is_child_hit(child, hit_obj):
                return True
        return False


@gin.configurable
def is_visible(
    loc: mathutils.Vector,
    obj_to_hit: bpy.types.Object,
    camera: bpy.types.Camera = None,
) -> bool:
    """ Cast a ray to determine if object is visible from camera. """
    scene = zpy.blender.verify_blender_scene()
    if camera is None:
        camera = scene.camera
    view_layer = zpy.blender.verify_view_layer()
    result = scene.ray_cast(depsgraph=view_layer.depsgraph,
                            origin=camera.location,
                            direction=(loc - camera.location))
    # Whether a hit occured
    is_hit = result[0]
    # Object hit by raycast
    hit_obj = result[4]
    if not is_hit:
        # Nothing was hit by the ray
        log.debug(f'No raycast hit from camera to {obj_to_hit.name}')
        return False
    if is_child_hit(obj_to_hit, hit_obj):
        # One of the children of the obj_to_hit was hit
        log.debug(f'Raycast hit from camera to {obj_to_hit.name}')
        return True
    return False


@gin.configurable
def is_in_view(
    loc: mathutils.Vector,
    camera: bpy.types.Camera = None,
    epsilon: float = 0.05,
) -> bool:
    """ Is a point visible to camera? Within some epsilon. """
    x, y, z = camera_xyz(loc, camera=camera)
    if z < 0:
        return False
    if x < (0-epsilon) or x > (1 + epsilon):
        return False
    if y < (0-epsilon) or y > (1 + epsilon):
        return False
    return True


@gin.configurable
def camera_xyv(
    loc: mathutils.Vector,
    obj: bpy.types.Object,
    camera: bpy.types.Camera = None,
    image_width: int = 640,
    image_height: int = 480,
) -> Tuple[int]:
    """ Get camera image xyv coordinates of point in scene.

    Keypoint coordinates (x, y) are measured from the top left
    image corner (and are 0-indexed). Coordinates are rounded
    to the nearest pixel to reduce file size. Visibility (v):

        v=0: not labeled (in which case x=y=0)
        v=1: labeled but not visible
        v=2: labeled and visible

    """
    x, y, z = camera_xyz(loc, camera=camera)
    # visibility
    v = 2
    if x < 0 or y < 0 or z < 0:
        v = 1
    if not is_visible(loc, obj_to_hit=obj, camera=camera):
        v = 1
    # bottom-left to top-left
    y = 1 - y
    # float (0, 1) to pixel int (0, pixel size)
    x = int(x * image_width)
    y = int(y * image_height)
    return x, y, v
